'''
--------------------------------------------------------------------------------

    flea2.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea

class Flea2(Lea):
    
    '''
    Flea2 is a Lea subclass, which instance is defined by a function applied on two given arguments.
    The function is applied on all elements of cartesian product of the arguments. This results in a new
    probability distribution for all the values returned by the function.
    '''
    
    __slots__ = ('_f','_leaArg1','_leaArg2')

    def __init__(self,f,arg1,arg2):
        Lea.__init__(self)
        self._f = f
        self._leaArg1 = Lea.coerce(arg1)
        self._leaArg2 = Lea.coerce(arg2)

    def _getLeaChildren(self):
        return (self._leaArg1,self._leaArg2)

    def _clone(self,cloneTable):
        return Flea2(self._f,self._leaArg1.clone(cloneTable),self._leaArg2.clone(cloneTable))    

    def _genVPs(self):
        f = self._f
        for (v1,p1) in self._leaArg1._genVPs():
            for (v2,p2) in self._leaArg2._genVPs():
                yield (f(v1,v2),p1*p2)

    def _genOneRandomMC(self):
        for v1 in self._leaArg1._genOneRandomMC():
            for v2 in self._leaArg2._genOneRandomMC():
                yield self._f(v1,v2)



'''
--------------------------------------------------------------------------------

    flea2.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from flea2 import Flea2

class Flea2a(Flea2):
    
    '''
    Flea2a is a Flea2 subclass, which instance is defined by a function applied on two given Lea arguments
    with a given "right-absorber" value (i.e. f(x,absorber) = absorber). This gives equivalent results as
    Flea2 (without absorber) but these could be more efficient by pruning the tree search. 
    The function is applied on all elements of cartesian product of the arguments. This results in a new
    probability distribution for all the values returned by the function.
    '''
    
    __slots__ = ('_absorber',)

    def __init__(self,f,arg1,arg2,absorber):
        Flea2.__init__(self,f,arg1,arg2)
        self._absorber = absorber

    def _clone(self,cloneTable):
        return Flea2a(self._f,self._leaArg1.clone(cloneTable),self._leaArg2.clone(cloneTable),self._absorber)    

    def _genVPs(self):
        f = self._f
        absorber = self._absorber
        for (v2,p2) in self._leaArg2._genVPs():
            if v2 is absorber:
                yield (absorber,self._leaArg1._getCount()*p2)
            else:
                for (v1,p1) in self._leaArg1._genVPs():
                    yield (f(v1,v2),p1*p2)



'''
--------------------------------------------------------------------------------

    olea.py

--------------------------------------------------------------------------------
Copyright 2013 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from flea import Flea

class Olea(Flea):
    
    '''
    Olea is a Flea subclass, which instance represents a joint probability distribution.
    An Olea instance is built from a sequence of n given attribute names and a given Lea
    instance with n-tuples as values; the Olea instance carries an inner "data" class built
    from the given attribute names and each value of the Olea instance is an instance
    of this class.
    '''

    __slots__ = ('_lea1','_attrNames','_class')

    def __init__(self,attrNames,lea1):
        ''' each value of lea1 is a tuple having same cardinality
            as attrNames
        '''
        self._attrNames = attrNames
        self._buildClass()
        Flea.__init__(self,self._class,lea1)

    def _clone(self,cloneTable):
        return Olea(self._attrNames,self._cleaArgs.clone(cloneTable))

    def _buildClass(self):
        classAttrDict = dict(('__maxLength'+attrName,0) for attrName in self._attrNames)
        classAttrDict['__slots__'] = tuple(self._attrNames)
        self._class = type('',(_TemplateClass,),classAttrDict)


class _TemplateClass(object):
    
    def __init__(self,*args):
        object.__init__(self)
        aClass = self.__class__
        for (attrName,arg) in zip(aClass.__slots__,args):
            setattr(self,attrName,arg)
            length = len(str(arg))
            maxLengthAttrName = '__maxLength' + attrName
            if length > getattr(aClass,maxLengthAttrName):
                setattr(aClass,maxLengthAttrName,length)
                aClass.__templateStr = "<%s>" % ', '.join("%s=%%%ds"%(attrName,getattr(aClass,'__maxLength'+attrName)) for attrName in aClass.__slots__)

    def __str__(self):
        aClass = self.__class__
        return aClass.__templateStr % tuple(getattr(self,attrName) for attrName in aClass.__slots__)
    __repr__ = __str__

    def __hash__(self):
        return hash(tuple(getattr(self,attrName) for attrName in self.__class__.__slots__))
    
    def __lt__(self, other):
        return _TemplateClass.__cmp__(self,other) == -1

    # Python 2 compatibility    
    def __cmp__(self,other):
        for attrName in self.__class__.__slots__:
            v1 = getattr(self,attrName)
            v2 = getattr(other,attrName)
            res = (v1 > v2) - (v1 < v2)
            if res != 0:
                break
        return res



'''
--------------------------------------------------------------------------------

    alea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea
from prob_fraction import ProbFraction
from random import randrange
from bisect import bisect_left, bisect_right
from math import log, sqrt, exp
from toolbox import LOG2, memoize, zip, next, dict, defaultdict
import operator
import sys


class Alea(Lea):
    
    '''
    Alea is a Lea subclass, which instance is defined by explicit probability distribution data.
    An Alea instance is defined by given value-probability pairs. Each probability is
    defined as a positive "counter" or "weight" integer, without upper limit. The actual
    probabilities are calculated by dividing the counters by the sum of all counters.
    Values having null probability counters are dropped.
    '''

    __slots__ = ('_vs','_ps','_count','_val','_cumul','_invCumul','_randomIter','_cachesByFunc')
    
    def __init__(self,vs,ps):
        ''' initializes Alea instance's attributes
            vs is a sequence of values
            ps is a sequence of probability weights (same size and order as ps)
        '''
        Lea.__init__(self)
        # for an Alea instance, the alea cache is itself
        self._alea = self
        self._vs = vs
        self._ps = ps
        self._count = sum(ps)
        # _val is the value temporarily bound to the instance, during evaluation (see _genVPs method)
        # note: self is used as a sentinel value to express that no value is currently bound; Python's
        # None is not a good sentinel value since it prevents using None as value in a distribution 
        self._val = self
        self._cumul = [0]
        self._invCumul = []
        self._randomIter = self._createRandomIter()
        self._cachesByFunc = dict()

    # constructor methods
    # -------------------
    
    def new(self):
        ''' returns a new Alea instance which is an independent clone of self;
            note that the present method overloads Lea.new to be more efficient
        '''
        # note that the new Alea instance shares the immutable _vs and _ps attributes of self
        newAlea = Alea(self._vs,self._ps)
        # it can share also the mutable _cumul and _invCumul attributes of self (lists)
        newAlea._cumul = self._cumul
        newAlea._invCumul = self._invCumul
        return newAlea

    @staticmethod
    def fromValFreqsDictGen(probDict):
        ''' static method, returns an Alea instance representing a distribution
            for the given probDict dictionary of {val:prob}, where prob is an integer number,
            a floating-point number or a fraction (Fraction or ProbFraction instance)
            so that each value val has probability proportional to prob to occur
            any value with null probability is ignored (hence not stored)
            if the sequence is empty, then an exception is raised
        '''
        probFractions = tuple(ProbFraction.coerce(probWeight) for probWeight in probDict.values())
        # TODO Check positive
        probWeights = ProbFraction.getProbWeights(probFractions)
        return Alea.fromValFreqsDict(dict(zip(probDict.keys(),probWeights)))
    
    @staticmethod
    def fromValFreqsDict(probDict,reducing=True):
        ''' static method, returns an Alea instance representing a distribution
            for the given probDict dictionary of {val:prob}, where prob is an integer number
            so that each value val has probability proportional to prob to occur
            any value with null probability is ignored (hence not stored)
            if reducing is True, then the probability weights are reduced to have a GCD = 1
            if the sequence is empty, then an exception is raised
        '''
        count = sum(probDict.values())
        if count == 0:
            raise Lea.Error("impossible to build a probability distribution with no value")
        gcd = count
        impossibleValues = []
        # check probabilities, remove null probabilities and calculate GCD
        for (v,p) in probDict.items():
            if p < 0:
                raise Lea.Error("negative probability")
            if p == 0:
                impossibleValues.append(v)
            elif reducing and gcd > 1:
                while p != 0:
                    if gcd > p:
                        (gcd,p) = (p,gcd)
                    p %= gcd
        for impossibleValue in impossibleValues:
            del probDict[impossibleValue]
        vpsIter = probDict.items()
        if reducing:
            vpsIter = ((v,p//gcd) for (v,p) in vpsIter)
        vps = list(vpsIter)
        try:            
            vps.sort()
        except:
            # no ordering relationship on values (e.g. complex numbers)
            pass
        return Alea(*zip(*vps))

    @staticmethod
    def fromVals(*values):
        ''' static method, returns an Alea instance representing a distribution
            for the given sequence of values, so that each value occurrence is
            taken as equiprobable;
            if each value occurs exactly once, then the distribution is uniform,
            i.e. the probability of each value is equal to 1 / #values;
            if the sequence is empty, then an exception is raised
        '''
        probDict = dict()
        for value in values:
            probDict[value] = probDict.get(value,0) + 1
        return Alea.fromValFreqsDict(probDict)

    @staticmethod
    def _fromValFreqs(valueFreqs,reducing):
        ''' static method, returns an Alea instance representing a distribution
            for the given sequence of (val,freq) tuples, where freq is a natural number
            so that each value is taken with the given frequency (or sum of 
            frequencies of that value if it occurs multiple times);
            if reducing is True, then the frequencies are reduced by dividing them by
            their GCD
            if the sequence is empty, then an exception is raised
        '''        
        probDict = dict()
        for (value,freq) in valueFreqs:
            probDict[value] = probDict.get(value,0) + freq
        return Alea.fromValFreqsDict(probDict,reducing)

    @staticmethod
    def fromValFreqs(*valueFreqs):
        ''' static method, returns an Alea instance representing a distribution
            for the given sequence of (val,freq) tuples, where freq is a natural number
            so that each value is taken with the given frequency (or sum of 
            frequencies of that value if it occurs multiple times);
            the frequencies are reduced by dividing them by their GCD
            if the sequence is empty, then an exception is raised
        '''        
        return Alea._fromValFreqs(valueFreqs,True)

    @staticmethod
    def fromValFreqsNR(*valueFreqs):
        ''' static method, returns an Alea instance representing a distribution
            for the given sequence of (val,freq) tuples, where freq is a natural number
            so that each value is taken with the given frequency (or sum of 
            frequencies of that value if it occurs multiple times);
            if the sequence is empty, then an exception is raised
        '''
        return Alea._fromValFreqs(valueFreqs,False)

    @staticmethod
    def poisson(mean,precision):
        ''' static method, returns an Alea instance representing a Poisson probability
            distribution having the given mean; the distribution is approximated by
            the finite set of values that have probability > precision
            (i.e. low/high values with too small probabilities are dropped)
        '''
        precFactor = 0.5 / precision
        valFreqs = []
        p = exp(-mean)
        v = 0
        t = 0.
        while p > 0.0:
            valFreqs.append((v,int(0.5+p*precFactor)))
            t += p
            v += 1
            p = (p*mean) / v
        return Alea.fromValFreqs(*valFreqs)

    def asString(self,kind='/',nbDecimals=6,histoSize=100):
        ''' returns a string representation of probability distribution self;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value with its
            probability in a format depending of given kind, which is string among
            '/', '.', '%', '-', '/-', '.-', '%-'; 
            the probabilities are displayed as
            - if kind[0] is '/' : rational numbers "n/d" or "0" or "1"
            - if kind[0] is '.' : decimals with given nbDecimals digits
            - if kind[0] is '%' : percentage decimals with given nbDecimals digits
            - if kind[0] is '-' : histogram bar made up of repeated '-', such that
                                  a bar length of histoSize represents a probability 1 
            if kind[1] is '-', the histogram bars with '-' are appended after 
                               numerical representation of probabilities
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used
        '''
        if kind not in ('/', '.', '%', '-', '/-', '.-', '%-'):
            raise Lea.Error("invalid display format '%s'"%kind)
        valueStrings = tuple(str(v) for v in self._vs)
        ps = self._ps
        vm = max(len(v) for v in valueStrings)
        linesIter = (v.rjust(vm)+' : ' for v in valueStrings)
        probRepresentation = kind[0]
        withHisto = kind[-1] == '-'
        if probRepresentation == '/':
            pStrings = tuple(str(p) for p in ps)
            pm = len(str(max(p for p in ps)))
            if self._count == 1:
                den = ''
            else:
                den = '/%d' % self._count
            linesIter = (line+pString.rjust(pm)+den for (line,pString) in zip(linesIter,pStrings))
        else:
            c = float(self._count)    
            if probRepresentation == '.':
                fmt = "%%s%%.%df" % nbDecimals
                linesIter = (fmt%(line,p/c) for (line,p) in zip(linesIter,ps))
            elif probRepresentation == '%':
                fmt = "%%s%%%d.%df %%%%" % (4+nbDecimals,nbDecimals)
                linesIter = (fmt%(line,100.*p/c) for (line,p) in zip(linesIter,ps))    
        if withHisto:
            c = float(self._count)    
            linesIter = (line+' '+int(0.5+(p/c)*histoSize)*'-' for (line,p) in zip(linesIter,ps))
        return '\n'.join(linesIter)

    def __str__(self):
        ''' returns a string representation of probability distribution self;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value  with its
            probability expressed as a rational number "n/d" or "0" or "1";
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used;
            called on evalution of "str(self)" and "repr(self)"
        '''
        return self.asString()
          
    def asFloat(self,nbDecimals=6):
        ''' returns a string representation of probability distribution self;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value with its
            probability expressed as decimal with given nbDecimals digits;
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used;
        '''
        return self.asString('.',nbDecimals)
        
    def asPct(self,nbDecimals=1):
        ''' returns a string representation of probability distribution self;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value with its
            probability expressed as percentage with given nbDecimals digits;
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used;
        '''
        return self.asString('%',nbDecimals)

    def histo(self,size=100):
        ''' returns a string representation of probability distribution self;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value with its
            probability expressed as a histogram bar made up of repeated '-',
            such that a bar length of given size represents a probability 1
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used;
        '''
        return self.asString('-',histoSize=size)
                
    def getAleaLeavesSet(self):
        ''' returns a set containing all the Alea leaves in the tree having the root self
            in the present case of Alea instance, it returns the singleton set with self as element
        '''
        return frozenset((self,))        
     
    def _getLeaChildren(self):
        # Alea instance has no children
        return ()

    def _clone(self,cloneTable):
        # note that the new Alea instance shares the immutable _vs and _ps attributes of self
        return Alea(self._vs,self._ps)

    def _getCount(self):
        ''' returns the total probability weight count (integer) of current Alea;
            this value depends on current binding (hence, the calculated value cannot be cached)
        '''
        if self._val is self:
            return self._count
        return 1

    def _genVPs(self):
        ''' generates tuple (v,p) where v is a value of the current probability distribution
            and p is the associated probability weight (integer > 0);
            this obeys the "binding" mechanism, so if the same variable is refered multiple times in
            a given expression, then same value will be yielded at each occurrence; 
            "Statues" algorithm: 
            before yielding a value v, this value v is bound to the current instance;
            then, if the current calculation requires to get again values on the current
            instance, then the bound value is yielded with probability 1;
            the instance is rebound to a new value at each iteration, as soon as the execution
            is resumed after the yield;
            it is unbound at the end;
            the method calls the _genVPs method implemented in Lea subclasses;
        '''
        if self._val is not self:
            # distribution already bound to a value because genVPs has been called already on self
            # it is yielded as a certain distribution (unique yield)
            yield (self._val,1)
        else:
            # distribution not yet bound to a value
            try:
                # browse all (v,p) tuples 
                for (v,p) in zip(self._vs,self._ps):
                    # bind value v: this is important if an object calls genVPs on the same instance
                    # before resuming the present generator (see above)
                    self._val = v
                    # yield the bound value v with probability weight p
                    yield (v,p)
            finally:
                # unbind value v, after all values have been bound or if an exception has been raised
                self._val = self
         
    def _genOneRandomMC(self):
        ''' generates one random value from the current probability distribution,
            WITHOUT precalculating the exact probability distribution (contrarily to 'random' method);
            this obeys the "binding" mechanism, so if the same variable is refered multiple times in
            a given expression, then same value will be yielded at each occurrence; 
            before yielding the random value v, this value v is bound to the current instance;
            then, if the current calculation requires to get again a random value on the current
            instance, then the bound value is yielded;
            the instance is rebound to a new value at each iteration, as soon as the execution
            is resumed after the yield;
            the instance is unbound at the end;
            the method calls the _genOneRandomMC method implemented in Lea subclasses;
        '''
        if self._val is not self:
            # distribution already bound to a value, because genOneRandomMC has been called already on self 
            # yield the bound value, in order to be consistent
            yield self._val
        else:
            try:
                # bind value v: this is important if an object calls genOneRandomMC on the same 
                # instance before resuming the present generator (see above)
                self._val = self.randomVal()
                # yield the bound value v
                yield self._val
            finally:
                # unbind value, after the random value has been bound or if an exception has been raised
                self._val = self
    
    def _p(self,val,checkValType=False):
        ''' returns the probability p/s of the given value val, as a tuple of naturals (p,s)
            where s is the sum of the probability weights of all values 
                  p is the probability weight of the given value val (from 0 to s)
            note: the ratio p/s is not reduced
            if checkValType is True, then raises an exception if some value in the
            distribution has a type different from val's
        '''
        p1 = 0
        if checkValType:
            errVal = self  # dumy value
            typeToCheck = type(val)
        # note: shall not exit the loop by a break/return (unbinding)
        for (v,p) in self._genVPs():
            if checkValType and not isinstance(v,typeToCheck):
                errVal = v
            if p1 == 0 and v == val:
                p1 = p
        if checkValType and errVal is not self:
            raise Lea.Error("found <%s> value although <%s> is expected"%(type(errVal).__name__,typeToCheck.__name__))
        return (p1,self._count)

    def cumul(self):
        ''' returns a list with the probability weights p that self <= value ;
            there is one element more than number of values; the first element is 0, then
            the sequence follows the order defined on values; if an order relationship is defined
            on values, then the tuples follows their increasing order; otherwise, an arbitrary
            order is used, fixed from call to call
            Note : the returned list is cached
        '''
        if len(self._cumul) == 1:
            cumulList = self._cumul
            pSum = 0
            for p in self._ps:
                pSum += p
                cumulList.append(pSum)
        return self._cumul

    def invCumul(self):
        ''' returns a tuple with the probability weights p that self >= value ;
            there is one element more than number of values; the first element is 0, then
            the sequence follows the order defined on values; if an order relationship is defined
            on values, then the tuples follows their increasing order; otherwise, an arbitrary
            order is used, fixed from call to call
            Note : the returned list is cached
        '''
        if len(self._invCumul) == 0:
            invCumulList = self._invCumul
            pSum = self._count
            for p in self._ps:
                invCumulList.append(pSum)
                pSum -= p
            invCumulList.append(0)
        return self._invCumul
            
    def randomVal(self):
        ''' returns a random value among the values of self, according to their probabilities
        '''
        return next(self._randomIter)
        
    def _createRandomIter(self):
        ''' generates an infinite sequence of random values among the values of self,
            according to their probabilities
        '''
        count = self._count
        probs = self.cumul()[1:]
        vals = self._vs
        while True:
            yield vals[bisect_right(probs,randrange(count))]    
        
    def randomDraw(self,n=None,sorted=False):
        ''' if n is None, returns a tuple with all the values of the distribution,
            in a random order respecting the probabilities
            (the higher count of a value, the most likely the value will be in the
             beginning of the sequence)
            if n > 0, then only n different values will be drawn
            if sorted is True, then the returned tuple is sorted
        '''
        if n is None:
           n = len(self._vs)
        elif n < 0:
            raise Lea.Error("randomDraw method requires a positive integer")    
        if n == 0:
            return ()
        lea1 = self
        res = []
        while True:
            lea1 = lea1.getAlea()
            x = lea1.random()
            res.append(x)
            n -= 1
            if n == 0:
                break
            lea1 = lea1.given(lea1!=x)
        if sorted:
            res.sort()
        return tuple(res)
    
    @memoize
    def pCumul(self,val):
        ''' returns, as an integer, the probability weight that self <= val
            note that it is not required that val is in the support of self
        '''
        return self.cumul()[bisect_right(self._vs,val)] 

    @memoize
    def pInvCumul(self,val):
        ''' returns, as an integer, the probability weight that self >= val
            note that it is not required that val is in the support of self
        '''
        return self.invCumul()[bisect_left(self._vs,val)] 

    @staticmethod
    def fastExtremum(cumulFunc,*aleaArgs):
        ''' static method, returns a new Alea instance giving the probabilities
            to have the extremum value (min or max) of each combination of the
            given Alea args;
            cumulFunc is the cumul function that determines whether max or min is
            used : respectively, Alea.pCumul or Alea.pInvCumul;
            the method uses an efficient algorithm (linear complexity), which is
            due to Nicky van Foreest; for explanations, see
            http://nicky.vanforeest.com/scheduling/cpm/stochasticMakespan.html
        '''
        if len(aleaArgs) == 1:
            return aleaArgs[0]
        if len(aleaArgs) == 2:
            (aleaArg1,aleaArg2) = aleaArgs
            valFreqsDict = defaultdict(int)
            for (v,p) in aleaArg1._genVPs():
                valFreqsDict[v] = p * cumulFunc(aleaArg2,v)
            for (v,p) in aleaArg2._genVPs():
                valFreqsDict[v] += (cumulFunc(aleaArg1,v)-aleaArg1._p(v)[0]) * p
            return Lea.fromValFreqsDict(valFreqsDict)
        return Alea.fastExtremum(cumulFunc,aleaArgs[0],Alea.fastExtremum(cumulFunc,*aleaArgs[1:]))

    # WARNING: the following methods are called without parentheses (see Lea.__getattr__)

    indicatorMethodNames = ('P','Pf','mean','var','std','mode','entropy','information')

    def P(self):
        ''' returns a ProbFraction instance representing the probability of True,
            from 0/1 to 1/1;
            raises an exception if some value in the distribution is not boolean
            (this is NOT the case with self.p(True))
            WARNING: this method is called without parentheses
        '''
        return ProbFraction(*self._p(True,checkValType=True))

    def Pf(self):
        ''' returns the probability of True, as a floating point number,
            from 0.0 to 1.0;
            raises an exception if some value in the distribution is not boolean
            (this is NOT the case with self.pmf(True))
            WARNING: this method is called without parentheses
        '''
        return float(self.P)
        
    def mean(self):
        ''' returns the mean value of the probability distribution, which is the
            probability weighted sum of the values;
            requires that
            1 - the values can be subtracted together,
            2 - the differences of values can be multiplied by integers,
            3 - the differences of values multiplied by integers can be
                added to the values,
            4 - the sum of values calculated in 3 can be divided by a float
                or an integer;
            if any of these conditions is not met, then the result depends of the
            value class implementation (likely, raised exception)
            WARNING: this method is called without parentheses
        '''
        res = None
        x0 = None
        for (x,p) in self._genVPs():
            if x0 is None:
                x0 = x
            elif res is None:
                res = p * (x-x0)
            else:
                res += p * (x-x0)
        if res is not None:
            try:
                x0 += res / float(self._count)
            except:
                # if the / operator is not supported with float as denominator
                # e.g. datetime.timedelta in Python 2.x 
                x0 += res / self._count    
        return x0
   
    def var(self):
        ''' returns a float number representing the variance of the probability distribution;
            requires that
            1 - the requirements of the mean() method are met,
            2 - the values can be subtracted to the mean value,
            3 - the differences between values and the mean value can be squared;
            if any of these conditions is not met, then the result depends of the
            value implementation (likely, raised exception)
            WARNING: this method is called without parentheses
        '''
        res = 0
        m = self.mean
        for (v,p) in self._genVPs():
            res += p*(v-m)**2
        return res / float(self._count)    

    def std(self):
        ''' returns a float number representing the standard deviation of the probability distribution
            requires that the requirements of the variance() method are met
            WARNING: this method is called without parentheses
        '''      
        return sqrt(self.var)
 
    def mode(self):
        ''' returns a tuple with the value(s) of the probability distribution having the highest probability 
            WARNING: this method is called without parentheses
        '''
        maxP = max(self._ps)
        return tuple(v for (v,p) in self._genVPs() if p == maxP)
            
    def entropy(self):
        ''' returns a float number representing the entropy of the probability  distribution
            WARNING: this method is called without parentheses
        '''
        res = 0
        count = float(self._count)
        for (v,p) in self._genVPs():
            if p > 0:
               p /= count
               res -= p*log(p)
        return res / LOG2



'''
--------------------------------------------------------------------------------

    clea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea
from toolbox import calcLCM


class Clea(Lea):
    
    '''
    Clea is a Lea subclass, which instance is defined by a given sequence (L1,...Ln)
    of Lea instances; it represents a probability distribution made up from the
    cartesian product L1 x ... x Ln; assuming independency of events, it associates
    each (v1,...,vn) tuple with probability product P1(v1)...Pn(vn).
    '''
    
    __slots__ = ('_leaArgs',)

    def __init__(self,*args):
        Lea.__init__(self)
        self._leaArgs = tuple(Lea.coerce(arg) for arg in args)

    def _getLeaChildren(self):
        return self._leaArgs
    
    def _clone(self,cloneTable):
        return Clea(*(leaArg.clone(cloneTable) for leaArg in self._leaArgs))

    @staticmethod
    def prod(gs):
        if len(gs) == 0:
            yield ()
        else:
            for xs in Clea.prod(gs[:-1]):
                for x in gs[-1]():
                    yield xs + (x,)

    def _genVPs(self):
        for vps in Clea.prod(tuple(leaArg._genVPs for leaArg in self._leaArgs)):
            v = tuple(v for (v,p) in vps)
            p = 1
            for (v1,p1) in vps:
                p *= p1
            yield (v,p)

    def _genOneRandomMC(self):
        for v in Clea.prod(tuple(leaArg._genOneRandomMC for leaArg in self._leaArgs)):
            yield v



'''
--------------------------------------------------------------------------------

    license.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

VER = "2.2.0-beta.2"

licenseText = '''\
Lea %s Copyright (C) 2013, 2014, 2015 Pierre Denis
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under certain conditions.
For more details, see COPYING and COPYING.LESSER files.''' % VER




'''
--------------------------------------------------------------------------------

    markov.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea 
from alea import Alea 
from blea import Blea
from toolbox import zip, dict
from itertools import islice, tee

class Chain(object):
    '''
    A Chain instance represents a Markov chain, with a given set of states
    and given probabilities of transition from state to state.    
    '''

    __slots__ = ('_stateObjs','_stateAleaDict','_state','_nextStateBlea')
    
    def __init__(self,nextStateLeaPerState):
        ''' initializes Chain instance's attributes; 
            nextStateLeaPerState is a sequence of tuples (stateObj,nextStateLea)
            where stateObj is a state object (e.g. a string) and nextStateLea is a Lea instance
            giving probabilities of transition from stateObj to each state object 
        '''
        object.__init__(self)
        self._stateObjs = tuple(stateObj for (stateObj,nextStateLea) in nextStateLeaPerState)
        self._stateAleaDict = dict((stateObj,StateAlea(Lea.coerce(stateObj),self)) for stateObj in self._stateObjs)
        self._state = StateAlea(Lea.fromVals(*self._stateObjs),self)
        iterNextStateData = ((self._state==stateObj,nextStateLea) for (stateObj,nextStateLea) in nextStateLeaPerState)
        self._nextStateBlea = Blea.build(*iterNextStateData)

    @staticmethod
    def fromMatrix(stateObjs,*transProbsPerState):
        ''' returns a new Chain instance from given arguments
            stateObjs is a sequence of objects representing states (strings, usually);
            transProbsPerState arguments contain the transiiton probability weights;
            there is one such argument per state, it is a tuple (stateObj,transProbs)
            where transProbs is the sequence of probability weights of transition from
            stateObj to each declared state, in the order of their declarations 
        '''
        nextStateLeas = (Alea.fromValFreqs(*zip(stateObjs,transProbs)) for (stateObj,transProbs) in transProbsPerState)
        nextStateLeaPerState = tuple(zip(stateObjs,nextStateLeas))
        return Chain(nextStateLeaPerState)

    @staticmethod
    def fromSeq(stateObjSeq):
        ''' returns a new Chain instance from given sequence of state objects
            the probabilities of state transitions are set according to transition
            frequencies in the given sequence 
        '''
        (fromStateObjIter,toStateObjIter) = tee(stateObjSeq);
        for _ in toStateObjIter:
            break
        nextStateObjsDict = dict()
        for (fromStateObj,toStateObj) in zip(fromStateObjIter,toStateObjIter):
            nextStateObjs = nextStateObjsDict.get(fromStateObj)
            if nextStateObjs is None:
                nextStateObjs = []
                nextStateObjsDict[fromStateObj] = nextStateObjs
            nextStateObjs.append(toStateObj)
        nextStateNameAndObjs = list(nextStateObjsDict.items())
        nextStateNameAndObjs.sort()
        nextStateLeaPerState = tuple((stateObj,Alea.fromVals(*nextStateObjs)) for (stateObj,nextStateObjs) in nextStateNameAndObjs)
        return Chain(nextStateLeaPerState)        

    def __str(self,formatFunc):
        ''' returns a string representation of the Markov chain
            with each state S followed by the indented representation of probability distribution
            of transition from S to next state
            formatFunc is a function to represent probability distribution, taking a Lea instance
            as argument and returning a string 
        '''
        nextStateLeas = (nextStateILea._lea1 for nextStateILea in self._nextStateBlea._iLeas)
        formattedNextStateLeas = ('  -> ' + nextStateLea.map(str) for nextStateLea in nextStateLeas)
        return '\n'.join('%s\n%s'%(stateObj,formatFunc(formattedNextStateLea)) for (stateObj,formattedNextStateLea) in zip(self._stateObjs,formattedNextStateLeas))

    def asFloat(self):
        ''' same as __str__ but the probabilities are given in decimal representation 
        '''        
        return self.__str(Lea.asFloat)

    def asPct(self):
        ''' same as __str__ but the probabilities are given in percentage representation 
        '''        
        return self.__str(Lea.asPct)

    def __str__(self):
        ''' returns a string representation of the Markov chain
            with each state S followed by the indented representation of probability distribution
            of transition from S to next state
            the probabilities are given in fraction representation 
        '''        
        return self.__str(Lea.__str__)
        
    __repr__ = __str__

    def getStates(self):
        ''' returns a tuple containing one StateAlea instance per state declared in the chain,
            in the order of their declaration; each instance represents a certain, unique, state
        ''' 
        #return tuple(StateAlea(Lea.coerce(stateObj),self) for stateObj in self._stateObjs)
        return tuple(self._stateAleaDict[stateObj] for stateObj in self._stateObjs)

    def getState(self,stateObjLea):
        ''' returns a StateAlea instance corresponding to the probability distribution
            given in stateObjLea
            if stateObjLea is not a Lea instance, then it is assumed to be a certain state
        '''
        if isinstance(stateObjLea,Lea):
            return StateAlea(stateObjLea,self)
        # stateObjLea is not Lea instance: assume that it is a certain state object
        return self._stateAleaDict[stateObjLea]

    def nextState(self,fromState=None,n=1):
        ''' returns the StateAlea instance obtained after n transitions from initial state
            defined by the given fromState, instance of StateAlea
            if fromState is None, then the initial state is the uniform distribution of the declared states
            if n = 0, then this initial state is returned
        '''
        if n < 0:
            raise Lea.Error("nextState method requires a positive value for argument 'n'")
        if fromState is None:
            fromState = self._state
        stateN = Lea.coerce(fromState).getAlea()
        while n > 0:
            n -= 1
            stateN = self._nextStateBlea.given(self._state==stateN).getAlea()
        return StateAlea(stateN,self)

    def stateGiven(self,condLea):
        ''' returns the StateAlea instance verifying the given cond Lea 
        '''
        return StateAlea(self._state.given(condLea),self)

    def nextStateGiven(self,condLea,n=1):
        ''' returns the StateAlea instance obtained after n transitions from initial state
            defined by the state distribution verifying the given cond Lea 
            if n = 0, then this initial state is returned
        '''
        fromState = self._state.given(condLea)
        return self.nextState(fromState,n)


class StateAlea(Alea):
    '''
    A StateAlea instance represents a probability distribution of states, for a given Markov chain
    '''
    
    __slots__ = ('_chain',)
    
    def __init__(self,stateObjLea,chain):
        ''' initializes StateAlea instance's attributes
            corresponding to the probability distribution given in stateObjLea
            and referring to the given chain, instance of Chain 
        '''
        Alea.__init__(self,*zip(*stateObjLea.getAlea()._genVPs()))
        self._chain = chain

    def nextState(self,n=1):
        ''' returns the StateAlea instance obtained after n transitions from initial state self
            if n = 0, then self is returned
        '''
        return self._chain.nextState(self,n)

    def genRandomSeq(self):
        ''' generates an infinite sequence of random state objects,
            starting from self and obeying the transition probabilities defined in the chain
        '''
        state = self
        while True:
            stateObj = state.nextState().random()
            yield stateObj
            state = self._chain.getState(stateObj)

    def randomSeq(self,n):
        ''' returns a tuple containing n state objects representing a random sequence
            starting from self and obeying the transition probabilities defined in the chain
        '''
        return tuple(islice(self.genRandomSeq(),n))



'''
--------------------------------------------------------------------------------

    lea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

import operator
from itertools import islice
from math import log
from collections import namedtuple
from prob_fraction import ProbFraction
from toolbox import calcGCD, log2, makeTuple, easyMin, easyMax, readCSVFilename, readCSVFile, dict, zip

class Lea(object):
    
    '''
    Lea is an abstract class representing discrete probability distributions.

    Each instance of concrete Lea's subclasses (called simply a "Lea instance" in the following)
    represents a discrete probability distribution, which associates each value of a set of
    values with the probability that such value occurs.

    A Lea instance can be defined by a sequence of (value,weight), giving the probability weight 
    of each value. Such probability weights are natural numbers. The actual probability of a
    given value can be calculated by dividing a weight by the sum of all weights. A Lea instance
    can be defined also by a sequence of values, their probability weight being their number of
    occurences in the sequence.

    Lea instances can be combined in arithmetic expressions resulting in new Lea instances, by
    obeying the following rules:

    - Lea instances can be added, subtracted, multiplied and divided together,
    through +, -, *, /, // operators. The resulting distribution's values and probabilities
    are determined by combination of operand's values with a sum weighted by probability
    products (the operation known as 'convolution', for the adition case).
    - Other supported binary arithmetic operators are power (**), modulo (%) and
    divmod function.
    - Unary operators +, - and abs function are supported also.
    - The Python's operator precedence rules, with the parenthesis overrules, are fully
    respected.
    - Any object X, which is not a Lea instance, involved as argument of an
    expression containing a Lea instance, is coerced to a Lea instance
    having X has sole value, with probability 1 (i.e. occurrence of X is certain).
    - Lea instances can be compared together, through ==, !=, <, <=, >, >= operators.
    The resulting distribution is a boolean distribution, giving probability of True result
    and complementary probability of False result.
    - Boolean distributions can be combined together with AND, OR, XOR, through &, |, ^
    operators, respectively.

    WARNING: the Python's and, or, not, operators shall NOT be used on Lea instances because
    they do not return any sensible result. Replace:
           a and b    by    a & b
           a or b     by    a | b
           not a      by    ~ a

    WARNING: in boolean expression involving arithmetic comparisons, the parenthesis
    shall be used, e.g. (a < b) & (b < c)

    WARNING: the augmented comparison (a < b < c) expression shall NOT be used.; it does
    not return any sensible result (reason: it has the same limitation as 'and' operator).

    Lea instances can be used to generate random values, respecting the given probabilities.
    There are two Lea methods for this purpose:
    - random:   calculates the exact probabiliy distribution, then takes random values 
    - randomMC: takes random values from atomic probabiliy distribution, then makes the 
                required calculations (Monte-Carlo algorithm)
    The randomMC is suited for complex distributions, when calculation of exact probability
    distribution is intractable. This could be used to provide an estimation of the probability
    distribution (see estimateMC method).

    There are ten concrete subclasses to Lea, namely:
      Alea, Clea, Plea, Flea, Flea1, Flea2, Tlea, Ilea, Rlea and Blea.
    
    Each subclass represents a "definition" of discrete probability distribution, with its own data
    or with references to other Lea instances to be combined together through a given operation.
    Each subclass defines what are the (value,probability) pairs or how they can be generated (see
    _genVPs method implemented in each Lea subclass). The Lea class acts as a facade, by providing
    different methods to instantiate these subclasses, so it is usually not needed to instantiate
    them explicitely. Here is an overview on these subclasses, with their relationships.

    - An Alea instance is defined by explicit value-probability pairs. Each probability is defined as
    a positive "counter" integer, without upper limit; the actual probability is calculated by
    dividing the counter by the sum of all counters; values having null probability counters are
    dropped.

    Instances of other Lea's subclasses represent probability distributions obtained by operations
    done on existing Lea instance(s). Any such instance forms a tree structure, having other Lea
    instances as nodes and Alea instances as leaves. This uses "lazy evaluation": actual (value,
    probability) pairs are calculated only at the time they are required (e.g. display, query 
    probability of a given value, etc); then, these are aggregated in a new Alea instance. This 
    Alea instance is then cached, as an attribute of the queried Lea instance, for speeding up next
    queries.
    
    Here is a brief presentation of these Lea's subclasses: 

    - Clea provides the cartesian product of a given sequence of Lea instances
    - Plea provides the cartesian product of one Lea instance with itself, a given number of times
    - Flea applies a given n-ary function to a given sequence of n Lea instances
    - Flea1 applies a given 1-ary function to a given Lea instance
    - Flea2 applies a given 2-ary function to two given Lea instances
    - Tlea applies a given 2-ary function, a given number of times, on a given Lea instance
    - Ilea filters the values of a given Lea instance according to a given Lea instance representing a boolean condition (conditional probabilities)
    - Rlea embeds Lea instances as values of a parent Lea instance 
    - Blea defines CPT, providing Lea instances corresponding to given conditions (used for bayesian networks)

    Note that Plea is meant to be a more efficient alternative to a Clea-based implementation;
    similarily, Flea1, Flea2, Tlea are more efficient alternatives to a Flea-based implementation.

    WARNING: The following methods are called without parentheses:
        mean, var, std, mode, entropy, information
    These are applicable on any Lea instance; these are implemented and documented in the Alea class.

    Short design notes:
    Lea uses the "template method" design pattern: the Lea base abstract class calls the following methods,
    which are implemented in each Lea's subclass: _clone, _getLeaChildren, _genVPs and _genOneRandomMC.
    Excepting the afore-mentioned estimateMC method, Lea performs EXACT calculation of probability distributions.
    It implements an original algorithm, called the "Statues" algorithm, by reference to the game of the same name;
    this uses a variable binding mechanism that relies on Python's generators. To learn more, see doc of
    Alea._genVPs method as well as other Xlea._genVPs methods implemented in Lea's subclasses. 
    '''

    class Error(Exception):
        ''' exception representing any violation of requirements of Lea methods  
        '''
        pass
        
    class _FailedRandomMC(Exception):
        ''' internal exception representing a failure to get a set of random values that
            satisfy a given condition in a given number of trials (see '...MC' methods) 
        '''
        pass

    # Lea attributes
    __slots__ = ('_alea',)

    # constructor methods
    # -------------------

    def __init__(self):
        ''' initializes Lea instance's attributes
        '''
        # alea instance acting as a cache when actual value-probability pairs have been calculated
        self._alea = None
        
    def _id(self):
        ''' returns a unique id, containing the concrete Lea class name as prefix
        '''
        return '%s#%s'%(self.__class__.__name__,id(self))

    def getAleaLeavesSet(self):
        ''' returns a set containing all the Alea leaves in the tree having the root self;
            this calls _getLeaChildren() method implemented in Lea's subclasses;
            this method is overloaded in Alea subclass to stop the recursion
        '''
        return frozenset(aleaLeaf for leaChild in self._getLeaChildren() for aleaLeaf in leaChild.getAleaLeavesSet())

    # constructor methods
    # -------------------
             
    def clone(self,cloneTable=None):
        ''' returns a deep copy of current Lea, without any value binding;
            if the Lea tree contains multiple references to the same Lea instance,
            then it is cloned only once and the references are copied in the cloned tree
            (the cloneTable dictionary serves this purpose);
            the method calls the _clone() method implemented in Lea's subclasses
        '''
        if cloneTable is None:
            cloneTable = dict()
        clonedLea = cloneTable.get(self)
        if clonedLea is None:
            clonedLea = self._clone(cloneTable)
            cloneTable[self] = clonedLea
            if self._alea is not None:
                clonedLea._alea = self._alea.clone(cloneTable)
        return clonedLea

    @staticmethod
    def fromVals(*vals):
        ''' static method, returns an Alea instance representing a distribution
            for the given values passed as arguments, so that each value
            occurrence is taken as equiprobable;
            if each value occurs exactly once, then the distribution is uniform,
            i.e. the probability of each value is equal to 1 / #values;
            if the sequence is empty, then an exception is raised
        '''
        return Alea.fromVals(*vals)

    @staticmethod
    def fromSeq(sequence):
        ''' static method, returns an Alea instance representing a distribution
            for the given sequence of values (e.g. a list, tuple, iterator,...),
            so that each value occurrence is taken as equiprobable;        
            if each value occurs exactly once, then the distribution is uniform,
            i.e. the probability of each value is equal to 1 / #values;
            if the sequence is empty, then an exception is raised
        '''
        return Lea.fromVals(*sequence)

    @staticmethod
    def fromValFreqs(*valFreqs):
        ''' static method, returns an Alea instance representing a distribution
            for the given sequence of (val,freq) tuples, where freq is a natural number
            so that each value is taken with the given frequency (or sum of 
            frequencies of that value if it occurs multiple times);
            the frequencies are reduced by dividing them by their GCD
            if the sequence is empty, then an exception is raised
        '''
        return Alea.fromValFreqs(*valFreqs)
    
    @staticmethod
    def fromValFreqsNR(*valFreqs):
        ''' static method, returns an Alea instance representing a distribution
            for the given sequence of (val,freq) tuples, where freq is a natural number
            so that each value is taken with the given frequency (or sum of 
            frequencies of that value if it occurs multiple times);
            if the sequence is empty, then an exception is raised
        '''
        return Alea.fromValFreqsNR(*valFreqs)

    @staticmethod
    def fromValFreqsDict(probDict):
        ''' static method, returns an Alea instance representing a distribution
            for the given dictionary of {val:prob}, where prob is an integer number
            so that each value val has probability proportional to prob to occur
            if the sequence is empty, then an exception is raised
        '''
        return Alea.fromValFreqsDict(probDict)

    @staticmethod
    def fromValFreqsDictArgs(**probDict):
        ''' static method, same as fromValFreqsDict, excepting that the dictionary
            is passed in a **kwargs style
        '''
        return Lea.fromValFreqsDict(probDict)

    @staticmethod
    def boolProb(pNum,pDen=None):
        ''' static method, returns an Alea instance representing a boolean
            distribution such that probability of True is pNum/pDen
            if pDen is None, then pNum expresses the probability as a Fraction
        '''
        if pDen is None:
            # pNum is expected to be a Fraction
            pDen = pNum.denominator
            pNum = pNum.numerator
        ProbFraction(pNum,pDen).check()
        return Alea.fromValFreqs((True,pNum),(False,pDen-pNum))

    @staticmethod
    def bernoulli(pNum,pDen=None):
        ''' static method, returns an Alea instance representing a bernoulli
            distribution giving 1 with probability pNum/pDen and 0 with
            complementary probability;
            if pDen is None, then pNum expresses the probability as a Fraction
        '''
        if pDen is None:
            # pNum is expected to be a Fraction
            pDen = pNum.denominator
            pNum = pNum.numerator
        ProbFraction(pNum,pDen).check()
        return Alea.fromValFreqs((1,pNum),(0,pDen-pNum))

    @staticmethod
    def binom(n,pNum,pDen=None):
        ''' static method, returns an Alea instance representing a binomial
            distribution giving the number of successes among a number n of 
            independent experiments, each having probability pNum/pDen of success;  
            if pDen is None, then pNum expresses the probability as a Fraction
        '''
        return Lea.bernoulli(pNum,pDen).times(n)

    @staticmethod
    def poisson(mean,precision=1e-20):
        ''' static method, returns an Alea instance representing a Poisson probability
            distribution having the given mean; the distribution is approximated by
            the finite set of values that have probability > precision (= 1e-20 by default)
            (i.e. low/high values with too small probabilities are dropped)
        '''
        return Alea.poisson(mean,precision)

    @staticmethod
    def interval(fromVal,toVal):
        ''' static method, returns an Alea instance representing a uniform probability
            distribution, for all the integers in the interval [fromVal,toVal]
        '''
        return Lea.fromVals(*range(fromVal,toVal+1))

    @staticmethod
    def fromCSVFilename(csvFilename,dialect='excel',**fmtparams):
        ''' static method, returns an Alea instance representing the joint probability
            distribution of the data read in the CSV file of the given csvFilename;
            it is similar to Lea.fromCSVFile method, except that it takes a filename
            instead of an open file (i.e. the method opens itself the file for reading);
            see Lea.fromCSVFile doc for more details
        '''
        (attrNames,dataFreq) = readCSVFilename(csvFilename,dialect,**fmtparams)
        return Lea.fromValFreqs(*dataFreq).asJoint(*attrNames)

    @staticmethod
    def fromCSVFile(csvFile,dialect='excel',**fmtparams):
        ''' static method, returns an Alea instance representing the joint probability
            distribution of the data read in the given CSV file;
            the arguments follow the same semantics as those of Python's csv.reader
            method, which supports different CSV formats;
            see doc in https://docs.python.org/2/library/csv.html
            the fields found in the first read row of the CSV file provide information
            on the attributes: each field is made up of a name, which shall be a valid
            identifier, followed by an optional 3-characters type code among  
              {b} -> boolean
              {i} -> integer
              {f} -> float
              {s} -> string
              {#} -> count
            for example, using the comma delimiter (default), the first row
            in the CSV file could be:
                name,age{i},heigth{f},married{b}
            the type code define the conversion to be applied to the fields read on the
            data lines; if the type code is missing for a given field, the type string is
            assumed for this field; if the read value is empty, then it is converted to
            Python's None, except if the type is string, then, the value is the empty string; 
            if the read value is not empty and cannot be parsed for the expected type, then
            an exception is raised; for the boolean type, the following values (case
            insensitive):
              '1', 't', 'true', 'y', 'yes' are interpreted as Python's True,
              '0', 'f', 'false', 'n', 'no' are interpreted as Python's False;
            the {#} code identifies a field that provides a count number of the row,
            representing the probability of the row or its frequency as a positive integer;
            such field is NOT included as attribute of the joint distribution; it is useful
            to define non-uniform probability distribution, as alternative to repeating the
            same row multiple times
        '''
        (attrNames,dataFreq) = readCSVFile(csvFile,dialect,**fmtparams)
        return Lea.fromValFreqs(*dataFreq).asJoint(*attrNames)

    @staticmethod
    def fromPandasDF(dataframe,indexColName=None):
        ''' static method, returns an Alea instance representing the joint probability
            distribution from the given pandas dataframe;
            the attribute names of the distribution are those of the column of the 
            given dataframe; the first field in each item of the dataframe is assumed
            to be the index; its treatment depends on given indexColName:
            if indexColName is None, then this index field is ignored
            otherwise, it is put in the joint distribution with indexColName as 
            attribute name
        '''
        # TODO: retrieve index_col in df, if not 0
        attrNames = tuple(dataframe.columns)
        if indexColName is None:
            valuesIter = (t[1:] for t in dataframe.itertuples())
            attrNames = dataframe.columns
        else:
            valuesIter = dataframe.itertuples()
            attrNames = (indexColName,) + attrNames
        return Lea.fromSeq(valuesIter).asJoint(*attrNames)

    def withProb(self,condLea,pNum,pDen=None):
        ''' returns a new Alea instance from current distribution,
            such that pNum/pDen is the probability that condLea is true
            if pDen is None, then pNum expresses the probability as a Fraction
        '''
        curCondLea = Lea.coerce(condLea)
        reqCondLea = Lea.boolProb(pNum,pDen)
        if reqCondLea.isTrue():
            lea1 = self.given(curCondLea)
        elif not reqCondLea.isFeasible():
            lea1 = self.given(~curCondLea)
        else:    
            lea1 = Blea.if_(reqCondLea,self.given(curCondLea).getAlea(),self.given(~curCondLea).getAlea())
        return lea1.getAlea()
        
    def withCondProb(self,condLea,givenCondLea,pNum,pDen):
        ''' [DEPRECATED: use Lea.revisedWithCPT instead]
            returns a new Alea instance from current distribution,
            such that pNum/pDen is the probability that condLea is true
            given that givenCondLea is True, under the constraint that
            the returned distribution keeps prior probabilities of condLea
            and givenCondLea unchanged
        '''
        if not (0 <= pNum <= pDen):
            raise Lea.Error("%d/%d is outside the probability range [0,1]"%(pNum,pDen))
        condLea = Lea.coerce(condLea)
        givenCondLea = Lea.coerce(givenCondLea)
        # max 2x2 distribution (True,True), (True,False), (False,True), (True,True)
        # prior joint probabilities, non null probability
        d = self.map(lambda v:(condLea.isTrue(),givenCondLea.isTrue())).getAlea()
        e = dict(d._genVPs())
        eTT = e.get((True,True),0)
        eFT = e.get((False,True),0)
        eTF = e.get((True,False),0)
        eFF = e.get((False,False),0)
        nCondLeaTrue = eTT + eTF
        nCondLeaFalse = eFT + eFF
        nGivenCondLeaTrue = eTT + eFT
        # new joint probabilities
        nTT = nGivenCondLeaTrue*pNum
        nFT = nGivenCondLeaTrue*(pDen-pNum)
        nTF = nCondLeaTrue*pDen - nTT
        nFF = nCondLeaFalse*pDen - nFT
        # feasibility checks
        if eTT == 0 and nTT > 0:
            raise Lea.Error("unfeasible: probability shall remain 0")
        if eFT == 0 and nFT > 0:
            raise Lea.Error("unfeasible: probability shall remain 1")
        if eTF == 0 and nTF > 0:
            raise Lea.Error("unfeasible: probability shall remain %d/%d"%(nCondLeaTrue,nGivenCondLeaTrue)) 
        if eFF == 0 and nFF > 0:
            msg = "unfeasible"
            if nGivenCondLeaTrue >= nCondLeaTrue:
                msg += ": probability shall remain %d/%d"%(nGivenCondLeaTrue-nCondLeaTrue,nGivenCondLeaTrue)
            raise Lea.Error(msg)
        if nTF < 0 or nFF < 0:
            pDenMin = nGivenCondLeaTrue
            pNumMin = max(0,nGivenCondLeaTrue-nCondLeaFalse)
            pDenMax = nGivenCondLeaTrue
            pNumMax = min(pDenMax,nCondLeaTrue)
            gMin = calcGCD(pNumMin,pDenMin)
            gMax = calcGCD(pNumMax,pDenMax)
            pNumMin //= gMin 
            pDenMin //= gMin 
            pNumMax //= gMax 
            pDenMax //= gMax
            raise Lea.Error("unfeasible: probability shall be in the range [%d/%d,%d/%d]"%(pNumMin,pDenMin,pNumMax,pDenMax))
        w = { (True  , True ) : nTT,
              (True  , False) : nTF,
              (False , True ) : nFT,
              (False , False) : nFF }
        m = 1
        for r in e.values():
            m *= r      
        # factors to be applied on current probabilities
        # depending on the truth value of (condLea,givenCondLea) on each value
        w2 = dict((cg,w[cg]*(m//ecg)) for (cg,ecg) in e.items())
        return Alea.fromValFreqs(*((v,p*w2[(condLea.isTrue(),givenCondLea.isTrue())]) for (v,p) in self._genVPs()))

    def given(self,*evidences):
        ''' returns a new Ilea instance representing the current distribution
            updated with the given evidences, which are each either a boolean or
            a Lea instance with boolean values; the values present in the returned
            distribution are those and only those compatible with the given AND of
            evidences;
            the resulting (value,probability) pairs are calculated when the
            returned Ilea instance is evaluated;
            an exception is raised if the evidences contain a non-boolean or
            if they are unfeasible
        '''
        return Ilea(self,(Lea.coerce(evidence) for evidence in evidences))    
    
    def times(self,n,op=operator.add):
        ''' returns a new Tlea instance representing the current distribution
            operated n times with itself, through the given binary operator
        '''
        return Tlea(op,self,n)

    def timesTuple(self,n):
        ''' returns a new Tlea instance with tuples of length n, containing
            the cartesian product of self with itslef repeated n times
        '''
        return self.map(makeTuple).times(n)

    def cprod(self,*args):
        ''' returns a new Clea instance, representing the cartesian product of all
            arguments (coerced to Lea instances), including self as first argument 
        '''
        return Clea(self,*args)

    def cprodTimes(self,nTimes):
        ''' returns a new Plea instance, representing the cartesian product of self
            with itself, iterated nTimes
        '''
        return Plea(self,nTimes)

    @staticmethod
    def reduce(op,args,absorber=None):
        ''' static method, returns a new Flea2 instance that join the given args with
            the given function op, from left to right;
            requires that op is a 2-ary function, accepting self's values as arguments;
            requires that args contains at least one element
            if absorber is not None, then it is considered as a "right-absorber" value
            (i.e. op(x,absorber) = absorber); this activates a more efficient algorithm
            which prunes the tree search as soon as the absorber is met.
        '''
        argsIter = iter(args)
        res = next(argsIter)
        if absorber is None:
            for arg in argsIter:
                res = Flea2(op,res,arg)
        else:
            for arg in argsIter:
                res = Flea2a(op,res,arg,absorber)
        return res

    def merge(self,*leaArgs):
        ''' returns a new Blea instance, representing the merge of self and given leaArgs, i.e.
                  P(v) = (P1(v) + ... + Pn(v)) / n
            where P(v)  is the probability of value v in the merge result 
                  Pi(v) is the probability of value v in ((self,)+leaArgs)[i]
        '''
        leas = (self,) + leaArgs
        lea = Lea.fromSeq(range(len(leas)))
        return Blea(*(Ilea(leaArg,(lea==i,)) for (i,leaArg) in enumerate(leas)))
    
    def map(self,f,args=()):
        ''' returns a new Flea instance representing the distribution obtained
            by applying the given function f, taking values of self distribution
            as first argument and given args tuple as following arguments (expanded);
            requires that f is a n-ary function with 1 <= n = len(args)+1 
            note: f can be also a Lea instance, with functions as values
        '''
        return Flea.build(f,(self,)+args)

    def mapSeq(self,f,args=()):
        ''' returns a new Flea instance representing the distribution obtained
            by applying the given function f on each element of each value
            of self distribution; if args is not empty, then it is expanded
            and added as f arguments
            requires that f is a n-ary function with 1 <= n = len(args)+1 
            requires that self's values are sequences
            returned distribution values are tuples
            note: f can be also a Lea instance, with functions as values
        '''
        return self.map(lambda v: tuple(f(e,*args) for e in v))

    def asJoint(self,*attrNames):
        ''' returns a new Alea instance representing a joint probability distribution
            from the current distribution supposed to have n-tuples as values,
            to be associated with the given n attribute names
        '''
        NTClass = namedtuple('_',attrNames)
        return self.map(lambda aTuple: NTClass(*aTuple)).getAlea()

    def draw(self,nbValues):
        ''' returns a new Lea instance representing a probability distribution of the
            sequences of values obtained by the given number of draws without
            replacement from the current distribution
            Note: if nbValues = 1, then a Flea instance is returned,
                  otherwise a Blea instance is returned 
        '''
        if nbValues <= 0:
            raise Lea.Error("draw method requires a strictly positive integer")
        if nbValues == 1:
            return self.map(makeTuple)
        vps = tuple(self._genVPs())
        return Blea(*(Ilea((v,)+Alea.fromValFreqs(*((v2,p2) for (v2,p2) in vps if v2 is not v)).draw(nbValues-1),(self==v,)) for (v,p) in vps))

    def flat(self):
        ''' assuming that self's values are themselves Lea instances,
            returns a new Rlea instance representing a probability distribution of
            inner values of these Lea instances  
        '''
        return Rlea(self)      
      
    @staticmethod
    def coerce(value):
        ''' static method, returns a Lea instance corresponding the given value:
            if the value is a Lea instance, then it is returned
            otherwise, an Alea instance is returned, with given value
            as unique (certain) value
        '''
        if not isinstance(value,Lea):
            return Alea((value,),(1,))
        return value

    def equiv(self,other):
        ''' returns True iff self and other represent the same probability distribution,
            i.e. they have the same probability for each of their value
            returns False otherwise
        '''
        other = Lea.coerce(other)
        # set(...) is used to avoid any dependency on the order of values
        res = frozenset(self.vps()) == frozenset(other.vps())
        if not res:
            # the previous test assumed that the instances have the same denominator
            # this is not the case if one of them has been created with fromValFreqsNR method
            # make an 'advanced' test, by insuring that both instances have the same denominator
            s = Alea.fromValFreqs(*self.vps())
            o = Alea.fromValFreqs(*other.vps())
            res = frozenset(s.vps()) == frozenset(o.vps()) 
        return res

    def p(self,val=None):
        ''' returns a ProbFraction instance representing the probability of given value val,
            from 0/1 to 1/1
            if val is None, then a tuple is returned with the probabilities of each value,
            in the same order as defined on values (call vals method to get this 
            ordered sequence)
        '''
        if val is None:
            count = self.getAlea()._count
            return tuple(ProbFraction(p,count) for p in self.ps())
        return ProbFraction(*self._p(val))
 
    def vps(self):
        ''' generates tuples (v,p) where v is a value of self
            and p is the associated probability weight (integer > 0);
            the sequence follows the order defined on values
        '''
        return self.getAlea()._genVPs()

    def vals(self):
        ''' returns a tuple with values of self
            the sequence follows the increasing order defined on values
            if order is undefined (e.g. complex numbers), then the order is
            arbitrary but fixed from call to call
        '''
        return self.getAlea()._vs

    def ps(self):
        ''' returns a tuple with probability weights (integer > 0) of self
            the sequence follows the increasing order defined on values
            if order is undefined (e.g. complex numbers), then the order is
            arbitrary but fixed from call to call
        '''
        return self.getAlea()._ps
        
    def support(self):
        ''' same as vals method
        '''
        return self.vals()
              
    def pmf(self,val=None):
        ''' probability mass function
            returns the probability of the given value val, as a floating point number
            from 0.0 to 1.0
            if val is None, then a tuple is returned with the probabilities of each value,
            in the same order as defined on values (call vals method to get this 
            ordered sequence)
        '''
        if val is None:
            count = float(self.getAlea()._count)
            return tuple(p/count for p in self.ps())
        (p,count) = self._p(val)
        return p / float(count)
        
    def cdf(self,val=None):
        ''' cumulative distribution function
            returns the probability that self's value is less or equal to the given value val,
            as a floating point number from 0.0 to 1.0
            if val is None, then a tuple is returned with the probabilities of each value,
            in the same order as defined on values (call vals method to get this 
            ordered sequence); the last probability is always 1.0 
        '''
        count = float(self.getAlea()._count)
        if val is None:
            return tuple(p/count for p in self.cumul()[1:])
        return self.getAlea().pCumul(val)/count

    def _p(self,val,checkValType=False):
        ''' returns the probability p/s of the given value val, as a tuple of naturals (p,s)
            where
            s is the sum of the probability weights of all values 
            p is the probability weight of the given value val (from 0 to s)
            note: the ratio p/s is not reduced
            if checkValType is True, then raises an exception if some value in the
            distribution has a type different from val's
        '''
        return self.getAlea()._p(val,checkValType)

    def isAnyOf(self,*values):
        ''' returns a boolean probability distribution
            indicating the probability that a value is any of the values passed as arguments
        '''
        return Flea1(lambda v: v in values,self)

    def isNoneOf(self,*values):
        ''' returns a boolean probability distribution
            indicating the probability that a value is none of the given values passed as arguments 
        '''
        return Flea1(lambda v: v not in values,self)

    @staticmethod
    def buildCPTFromDict(aCPTDict,priorLea=None):
        ''' static method, same as buildCPT, with clauses specified in the aCPTDict dictionary
            {condition:result}
        '''
        return Blea.build(*(aCPTDict.items()),priorLea=priorLea)

    @staticmethod
    def if_(condLea,thenLea,elseLea):
        ''' static method, an instance of Blea representing the conditional probability table
            giving thenLea  if condLea is true
                   elseLea  otherwise
            this is a convenience method equivalent to 
              Lea.buildCPT((condLea,thenLea),(None,elseLea))
        '''
        return Blea.build((condLea,thenLea),(None,elseLea))
        
    @staticmethod
    def buildCPT(*clauses,**kwargs):
        ''' static method, returns an instance of Blea representing the conditional
            probability table (e.g. a node in a Bayes network) from the given clauses;
            each clause is a tuple (condition,result)
            where condition is a boolean or a Lea boolean distribution
              and result is a value or Lea distribution representing the result
                   assuming that condition is true
            the conditions from all clauses shall be mutually exclusive
            if a clause contains None as condition, then it is considered as a 'else'
            condition;
            the method supports three optional named arguments:
             'priorLea', 'autoElse' and 'check';
            'priorLea' and 'autoElse' are mutually exclusive, they require the absence
            of an 'else' clause (otherwise, an exception is raised); 
            * if priorLea argument is specified, then the 'else' clause is calculated
            so that the priorLea is returned for the unconditional case
            * if autoElse argument is specified as True, then the 'else' clause is
            calculated so that a uniform probability distribution is returned for
            the condition cases not covered in given clauses (principle of indifference);
            the values are retrieved from the results found in given clauses
            * if check argument is specified as False, then NO checks are made on the given
            clauses (see below); this can significantly increase performances, as the 
            set of clauses or variables become large; 
            by default (check arg absent or set to True), checks are made on clause
            conditions to ensure that they form a partition:
              1) the clause conditions shall be mutually disjoint, i.e. no subset
                 of conditions shall be true together;
              2) if 'else' is missing and not calculated through 'priorLea' nor 'autoElse',
                 then the clause conditions shall cover all possible cases, i.e. ORing
                 them shall be certainly true;
            an exception is raised if any of such conditions is not verified;
        '''
        return Blea.build(*clauses,**kwargs)

    def revisedWithCPT(self,*clauses):
        ''' returns an instance of Blea representing the conditional probability table
            (e.g. a node in a Bayes network) from the given clauses;
            each clause is a tuple (condition,result)
            where condition is a boolean or a Lea boolean distribution
              and result is a value or Lea distribution representing the result
                   assuming that condition is true
            the conditions from all clauses shall be mutually exclusive
            no clause can contain None as condition
            the 'else' clause is calculated so that the returned Blea if no condition is
            given is self
        ''' 
        return Blea.build(*clauses,priorLea=self)
    
    def buildBNfromJoint(self,*bnDefinition):
        ''' returns a named tuple of Blea instances representing a Bayes network
            with variables stored in attributes A1, ... , An, assuming that self
            is a Lea joint probability distribution having, as values, named tuples
            with the same set of attributes A1, ... , An (such Lea instance is
            returned by asJoint method, for example);
            each argument of given bnDefinition represent a dependency relationship
            from a set of given variables to one given variable; this is expressed as
            a tuple (srcVarNames, tgtVarName) where srcVarNames is a sequence of
            attribute names (strings) identifying 'from' variables and tgtName is the
            attribute name (string) identifying the 'to' variable;
            the method builds up the 'to' variable of the BN as a CPT calculated from
            each combination of 'from' variables in the joint probability distribution:
            for each such combination C, the distribution of 'to' variable is calculated
            by marginalisation on the joint probability distribution given the C condition;
            possible missing combinations are covered in an 'else' clause on the CPT
            that is defined as a uniform distribution of the values of 'to' variable,
            which are found in the other clauses (principle of indifference);
            the variables that are never refered as 'to' variable are considered
            as independent in the BN and are calculated by unconditional marginalisation
            on the joint probability distribution;
            if a variable appears in more than one 'to' variable, then an exception is
            raised (error)
        '''
        jointAlea = self.getAlea()
        # retrieve the named tuple class from the first value of the joint distribution,
        NamedTuple = jointAlea._vs[0].__class__
        varsDict = dict((varName,self.__getattribute__(varName)) for varName in NamedTuple._fields)
        # all BN variables initialized as independent (maybe overwritten below, according to given relationships)
        varsBNDict = dict((varName,var.getAlea()) for (varName,var) in varsDict.items())
        for (srcVarNames,tgtVarName) in bnDefinition:
            if not isinstance(varsBNDict[tgtVarName],Alea):
                raise Lea.Error("'%s' is defined as target in more than one BN relationship"%tgtVarName)
            tgtVar = varsDict[tgtVarName]
            cprodSrcVars = Lea.cprod(*(varsDict[srcVarName] for srcVarName in srcVarNames))
            cprodSrcVarsBN = Lea.cprod(*(varsBNDict[srcVarName] for srcVarName in srcVarNames))
            ## Note: the following statement is functionnaly equivalent to statements after
            ## BUT it is far slower in case of missing conditions on the CPT ('else' clause)
            ## because it relies on the (too) generic 'autoElse' algorithm of Blea, which makes 
            ## a negated OR on all given conditions;
            ## varsBNDict[tgtVarName] = Blea.build(
            ##                *((cprodSrcVarsBN==cprodSrcVal,tgtVar.given(cprodSrcVars==cprodSrcVal).getAlea()) \
            ##                   for cprodSrcVal in cprodSrcVars.vals()), autoElse=True)
            # build CPT clauses (condition,result) from the joint probability distribution
            cprodSrcVals = cprodSrcVars.vals()
            clauses = tuple((cprodSrcVarsBN==cprodSrcVal,tgtVar.given(cprodSrcVars==cprodSrcVal).getAlea()) \
                             for cprodSrcVal in cprodSrcVals)
            # determine missing conditions in the CPT, if any
            allVals = Lea.cprod(*(varsDict[srcVarName].getAlea() for srcVarName in srcVarNames)).vals()
            missingVals = frozenset(allVals) - frozenset(cprodSrcVals)
            if len(missingVals) > 0:
                # there are missing conditions: add clauses with each of these conditions associating  
                # them with a uniform distribution built on the values found in results of other clauses
                # (principle of indifference)
                elseResult = Lea.fromVals(*frozenset(val for (cond,result) in clauses for val in result.vals()))
                clauses += tuple((cprodSrcVarsBN==missingVal,elseResult) for missingVal in missingVals)
                ## Note: the following statements are functionally equivalent to the statement above
                ## these aggregate missing condtions with an OR; the tests show similar performance,
                ## albeit a small decrease : that's why the former has been preferred;
                ## elseCond = Lea.reduce(operator.or_,(cprodSrcVarsBN==missingVal for missingVal in missingVals),True)
                ## clauses += ((elseCond,elseResult),)
            # overwrite the target BN variable (currently independent Alea instance), with a CPT built
            # up from the clauses determined from the joint probability distribution
            # the check is deactivated for the sake of performance; this is safe since, by construction,
            # the clauses conditions verify the "truth partioning" rules  
            varsBNDict[tgtVarName] = Blea.build(*clauses,check=False)
        # return the BN variables as attributes of a new named tuple having the same attributes as the
        # values found in self
        return NamedTuple(**varsBNDict)
    
    @staticmethod
    def makeVars(obj,tgtDict,prefix='',suffix=''):
        ''' retrieve attributes names A1, ... , An of obj and put associations 
            {V1 : obj.A1, ... , Vn : obj.An} in tgtDict dictionary
            where Vi is a variable name string built as prefix + Ai + suffix;
            obj is
            (a) either a named tuple with attributes A1, ... , An (as returned
            by buildBNfromJoint, for example)
            (b) or a Lea instances representing a joint probability distribution
            with the attributes A1, ... , An (such Lea instance is returned by
            asJoint method, for example);
            note: if the caller passes globals() as tgtDict, then the variables
            named Vi, refering to obj.Ai, shall be created in its scope, as
            a side-effect (this is the purpose of the method);
            warning: the method may silently overwrite caller's variables
        '''
        if isinstance(obj,Lea):
            # case (b)
            # retrieve the named tuple class from the first value of the joint distribution
            NamedTuple = obj.getAlea()._vs[0].__class__
        else:
            # case (a)
            NamedTuple = obj.__class__
        tgtDict.update((prefix+varName+suffix,obj.__getattribute__(varName)) for varName in NamedTuple._fields)       
    
    def __call__(self,*args):
        ''' returns a new Flea instance representing the probability distribution
            of values returned by invoking functions of current distribution on 
            given arguments (assuming that the values of current distribution are
            functions);
            called on evaluation of "self(*args)"
        '''
        return Flea.build(self,args)

    def __getitem__(self,index):
        ''' returns a new Flea instance representing the probability distribution
            obtained by indexing or slicing each value with index
            called on evaluation of "self[index]"
        '''
        return Flea2(operator.getitem,self,index)

    def __iter__(self):
        ''' returns the iterator returned by genVP()
            called on evaluation of "iter(self)", "tuple(self)", "list(self" 
                or on "for x in self"
        '''
        return self._genVPs()

    def __getattribute__(self,attrName):
        ''' returns the attribute with the given name in the current Lea instance;
            if the attribute name is a distribution indicator, then the distribution
            is evaluated and the indicator method is called; 
            if the attribute name is unknown as a Lea instance's attribute,
            then returns a Flea instance that shall retrieve the attibute in the
            values of current distribution; 
            called on evaluation of "self.attrName"
            WARNING: the following methods are called without parentheses:
                         mean, var, std, mode, entropy, information
                     these are applicable on any Lea instance
                     and these are documented in the Alea class
        '''
        try:
            if attrName in Alea.indicatorMethodNames:
                # indicator methods are called implicitely
                return object.__getattribute__(self.getAlea(),attrName)()
            # return Lea's instance attribute
            return object.__getattribute__(self,attrName)
        except AttributeError:
            # return new Lea made up of attributes of inner values
            return Flea2(getattr,self,attrName)

    @staticmethod
    def fastMax(*args):
        ''' static method, returns a new Alea instance giving the probabilities to
            have the maximum value of each combination of the given args;
            if some elements of args are not Lea instance, then these are coerced
            to an Lea instance with probability 1;
            the method uses an efficient algorithm (linear complexity), which is
            due to Nicky van Foreest; for explanations, see
            http://nicky.vanforeest.com/scheduling/cpm/stochasticMakespan.html
            Note: unlike most of Lea methods, the distribution returned by Lea.fastMax
            loses any dependency with given args; this could be important if some args
            appear in the same expression as Lea.max(...) but outside it, e.g.
            conditional probability expressions; this limitation can be avoided by
            using the Lea.max method; however, this last method can be
            prohibitively slower (exponential complexity)
        '''
        aleaArgs = tuple(Lea.coerce(arg).getAlea() for arg in args)
        return Alea.fastExtremum(Alea.pCumul,*aleaArgs)
    
    @staticmethod
    def fastMin(*args):
        ''' static method, returns a new Alea instance giving the probabilities to have
            the minimum value of each combination of the given args;
            if some elements of args are not Lea instances, then these are coerced
            to an Alea instance with probability 1;
            the method uses an efficient algorithm (linear complexity), which is
            due to Nicky van Foreest; for explanations, see
            http://nicky.vanforeest.com/scheduling/cpm/stochasticMakespan.html
            Note: unlike most of Lea methods, the distribution returned by Lea.fastMin
            loses any dependency with given args; this could be important if some args
            appear in the same expression as Lea.min(...) but outside it, e.g.
            conditional probability expressions; this limitation can be avoided by
            using the Lea.min method; however, this last method can be prohibitively
            slower (exponential complexity)
        '''
        aleaArgs = tuple(Lea.coerce(arg).getAlea() for arg in args)
        return Alea.fastExtremum(Alea.pInvCumul,*aleaArgs)
        
    @staticmethod
    def max(*args):
        ''' static method, returns a new Flea instance giving the probabilities to
            have the maximum value of each combination of the given args;
            if some elements of args are not Lea instances, then these are coerced
            to a Lea instance with probability 1;
            the returned distribution keeps dependencies with args but the 
            calculation could be prohibitively slow (exponential complexity);
            for a more efficient implemetation, assuming that dependencies are not
            needed, see Lea.fastMax method
        '''
        return Flea.build(easyMax,args)

    @staticmethod
    def min(*args):
        ''' static method, returns a new Flea instance giving the probabilities to
            have the minimum value of each combination of the given args;
            if some elements of args are not Lea instances, then these are coerced
            to a Lea instance with probability 1;
            the returned distribution keeps dependencies with args but the 
            calculation could be prohibitively slow (exponential complexity);
            for a more efficient implemetation, assuming that dependencies are not
            needed, see Lea.fastMin method
        '''
        return Flea.build(easyMin,args)

    def __lt__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            that the values of self are less than the values of other;
            called on evaluation of "self < other"
        '''
        return Flea2(operator.lt,self,other)

    def __le__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            that the values of self are less than or equal to the values of other;
            called on evaluation of "self <= other"
        '''
        return Flea2(operator.le,self,other)

    def __eq__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            that the values of self are equal to the values of other;
            called on evaluation of "self == other"
        '''
        return Flea2(operator.eq,self,other)

    def __hash__(self):
        return id(self)

    def __ne__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            that the values of self are different from the values of other;
            called on evaluation of "self != other"
        '''
        return Flea2(operator.ne,self,other)

    def __gt__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            that the values of self are greater than the values of other;
            called on evaluation of "self > other"
        '''
        return Flea2(operator.gt,self,other)

    def __ge__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            that the values of self are greater than or equal to the values of other;
            called on evaluation of "self >= other"
        '''
        return Flea2(operator.ge,self,other)
    
    def __add__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the addition of the values of self with the values of other;
            called on evaluation of "self + other"
        '''
        return Flea2(operator.add,self,other)

    def __radd__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the addition of the values of other with the values of self;
            called on evaluation of "other + self"
        '''
        return Flea2(operator.add,other,self)

    def __sub__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the subtraction of the values of other from the values of self;
            called on evaluation of "self - other"
        '''
        return Flea2(operator.sub,self,other)

    def __rsub__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the subtraction of the values of self from the values of other;
            called on evaluation of "other - self"
        '''
        return Flea2(operator.sub,other,self)

    def __pos__(self):
        ''' returns a Flea instance representing the probability distribution
            resulting from applying the unary positive operator on the values of self;
            called on evaluation of "+self"
        '''
        return Flea1(operator.pos,self)

    def __neg__(self):
        ''' returns a Flea instance representing the probability distribution
            resulting from negating the values of self;
            called on evaluation of "-self"
        '''
        return Flea1(operator.neg,self)

    def __mul__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the multiplication of the values of self by the values of other;
            called on evaluation of "self * other"
        '''
        return Flea2(operator.mul,self,other)

    def __rmul__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the multiplication of the values of other by the values of self;
            called on evaluation of "other * self"
        '''
        return Flea2(operator.mul,other,self)

    def __pow__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the powering the values of self with the values of other;
            called on evaluation of "self ** other"
        '''
        return Flea2(operator.pow,self,other)

    def __rpow__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the powering the values of other with the values of self;
            called on evaluation of "other ** self"
        '''
        return Flea2(operator.pow,other,self)

    def __truediv__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the division of the values of self by the values of other;
            called on evaluation of "self / other"
        '''
        return Flea2(operator.truediv,self,other)

    def __rtruediv__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the division of the values of other by the values of self;
            called on evaluation of "other / self"
        '''
        return Flea2(operator.truediv,other,self)

    def __floordiv__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the floor division of the values of self by the values of other;
            called on evaluation of "self // other"
        '''
        return Flea2(operator.floordiv,self,other)

    def __rfloordiv__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the floor division of the values of other by the values of self;
            called on evaluation of "other // self"
        '''
        return Flea2(operator.floordiv,other,self)

    # Python 2 compatibility
    __div__ = __truediv__
    __rdiv__ = __rtruediv__

    def __mod__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the modulus of the values of self with the values of other;
            called on evaluation of "self % other"
        '''
        return Flea2(operator.mod,self,other)

    def __rmod__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the modulus of the values of other with the values of self;
            called on evaluation of "other % self"
        '''
        return Flea2(operator.mod,other,self)

    def __divmod__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from applying the function divmod on the values of self and the values of other;
            called on evaluation of "divmod(self,other)"
        '''
        return Flea2(divmod,self,other)

    def __rdivmod__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from applying the function divmod on the values of other and the values of self;
            called on evaluation of "divmod(other,self)"
        '''
        return Flea2(divmod,other,self)

    def __floordiv__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the integer division of the values of self by the values of other;
            called on evaluation of "self // other"
        '''
        return Flea2(operator.floordiv,self,other)
    
    def __rfloordiv__(self,other):
        ''' returns a Flea instance representing the probability distribution
            resulting from the integer division of the values of other by the values of self;
            called on evaluation of "other // self"
        '''
        return Flea2(operator.floordiv,other,self)

    def __abs__(self):
        ''' returns a Flea instance representing the probability distribution
            resulting from applying the abs function on the values of self;
            called on evaluation of "abs(self)"
        '''
        return Flea1(abs,self)
    
    def __and__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            resulting from the locical AND between the values of self and the values of other;
            note that the method uses short-circuit evaluation when other's value is false
            called on evaluation of "self & other"
        '''
        return Flea2a(Lea._safeAnd,self,other,False)

    def __rand__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            resulting from the locical AND between the values of other and the values of self;
            note that the method uses short-circuit evaluation when self's value is false
            called on evaluation of "other & self"
        '''
        return Flea2a(Lea._safeAnd,other,self,False)

    def __or__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            resulting from the locical OR between the values of self and the values of other;
            note that the method uses short-circuit evaluation when other's value is true
            called on evaluation of "self | other"
        '''
        return Flea2a(Lea._safeOr,self,other,True)

    def __ror__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            resulting from the locical OR between the values of other and the values of self;
            note that the method uses short-circuit evaluation when self's value is true
            called on evaluation of "other | self"
        '''
        return Flea2a(Lea._safeOr,other,self,True)

    def __xor__(self,other):
        ''' returns a Flea instance representing the boolean probability distribution
            resulting from the locical XOR between the values of self and the values of other;
            called on evaluation of "self ^ other"
        '''
        return Flea2(Lea._safeXor,self,other)

    def __invert__(self):
        ''' returns a Flea instance representing the boolean probability distribution
            resulting from the locical NOT of the values self;
            called on evaluation of "~self"
        '''
        return Flea1(Lea._safeNot,self)

    def __nonzero__(self):
        ''' raises an exception telling that Lea instance cannot be evaluated as a boolean
            called on evaluation of "bool(self)"
        '''
        raise Lea.Error("Lea instance cannot be evaluated as a boolean (maybe due to a lack of parentheses)")

    @staticmethod
    def _checkBooleans(opMsg,*vals):
        ''' static method, raise an exception if any of vals arguments is not boolean;
            the exception messsage refers to the name of a logical operation given in the opMsg argument
        '''
        for val in vals:
            if not isinstance(val,bool):
                raise Lea.Error("non-boolean object involved in %s logical operation (maybe due to a lack of parentheses)"%opMsg) 

    @staticmethod
    def _safeAnd(a,b):
        ''' static method, returns a boolean, which is the logical AND of the given boolean arguments; 
            raises an exception if any of arguments is not boolean
        '''
        Lea._checkBooleans('AND',a,b)
        return operator.and_(a,b)

    @staticmethod
    def _safeOr(a,b):
        ''' static method, returns a boolean, which is the logical OR of the given boolean arguments; 
            raises an exception if any of arguments is not boolean
        '''
        Lea._checkBooleans('OR',a,b)
        return operator.or_(a,b)

    @staticmethod
    def _safeXor(a,b):
        ''' static method, returns a boolean, which is the logical XOR of the given boolean arguments; 
            raises an exception if any of arguments is not boolean
        '''
        Lea._checkBooleans('XOR',a,b)
        return operator.xor(a,b)

    @staticmethod
    def _safeNot(a):
        ''' static method, returns a boolean, which is the logical NOT of the given boolean argument; 
            raises an exception if the argument is not boolean
        '''
        Lea._checkBooleans('NOT',a)
        return operator.not_(a)    

    def _getCount(self):
        ''' returns the total probability weight count (integer) of current Lea;
            this value depends on current binding(s) on depending Alea leaves (hence, the
            calculated value cannot be cached);
            Note that the returned value could be obtained also by calling _genVPs()
            method and summing all the weights; however, the present method does that in
            a far more efficient way
        '''
        count = 1
        for aleaLeaf in self.getAleaLeavesSet():
            count *= aleaLeaf._getCount()
        return count

    def _getLeaChildren(self):
        ''' returns a tuple containing all the Lea instances children of the current Lea;
            Lea._getLeaChildren method is abstract: it is implemented in all Lea's subclasses
        '''
        raise NotImplementedError("missing method '%s._getLeaChildren(self)'"%(self.__class__.__name__))

    def _clone(self,cloneTable):
        ''' returns a deep copy of current Lea, without any value binding;
            if the Lea tree contains multiple references to the same Lea instance,
            then it is cloned only once and the references are copied in the cloned tree
            (the cloneTable dictionary serves this purpose);
            Lea._clone method is abstract: it is implemented in all Lea's subclasses
        '''
        raise NotImplementedError("missing method '%s._clone(self,cloneTable)'"%(self.__class__.__name__))
        
    def _genVPs(self):
        ''' generates tuple (v,p) where v is a value of the current probability distribution
            and p is the associated probability weight (integer > 0);
            this obeys the "binding" mechanism, so if the same variable is refered multiple times in
            a given expression, then same value will be yielded at each occurrence;
            Lea._genVPs method is abstract: it is implemented in all Lea's subclasses
        '''
        raise NotImplementedError("missing method '%s._genVPs(self)'"%(self.__class__.__name__))
         
    def _genOneRandomMC(self):
        ''' generates one random value from the current probability distribution,
            WITHOUT precalculating the exact probability distribution (contrarily to 'random' method);
            this obeys the "binding" mechanism, so if the same variable is refered multiple times in
            a given expression, then same value will be yielded at each occurrence; 
            before yielding the random value v, this value v is bound to the current instance;
            then, if the current calculation requires to get again a random value on the current
            instance, then the bound value is yielded;
            the instance is rebound to a new value at each iteration, as soon as the execution
            is resumed after the yield;
            the instance is unbound at the end;
            Lea._genOneRandomMC method is abstract: it is implemented in all Lea's subclasses
        '''
        raise NotImplementedError("missing method '%s._genOneRandomMC(self)'"%(self.__class__.__name__))
            
    def genRandomMC(self,n,nbTries=None):
        ''' generates n random value from the current probability distribution,
            without precalculating the exact probability distribution (contrarily to 'random' method);
            nbTries, if not None, defines the maximum number of trials in case a random value
            is incompatible with a condition; this happens only if the current Lea instance
            is (referring to) an Ilea or Blea instance, i.e. 'given' or 'buildCPT' methods;
            WARNING: if nbTries is None, any infeasible condition shall cause an infinite loop
        '''
        for _ in range(n):
            remainingNbTries = 1 if nbTries is None else nbTries
            v = self
            while remainingNbTries > 0:
                try:
                    for v in self._genOneRandomMC():
                        yield v
                    remainingNbTries = 0
                except Lea._FailedRandomMC:
                    if nbTries is not None:
                        remainingNbTries -= 1        
            if v is self:
                raise Lea.Error("impossible to validate given condition(s), after %d random trials"%nbTries) 
        
    def randomMC(self,n=None,nbTries=None):
        ''' if n is None, returns a random value with the probability given by the distribution
            without precalculating the exact probability distribution (contrarily to 'random' method);
            otherwise, returns a tuple of n such random values;
            nbTries, if not None, defines the maximum number of trials in case a random value
            is incompatible with a condition; this happens only if the current Lea instance
            is (referring to) an Ilea or Blea instance, i.e. 'given' or 'buildCPT' methods;
            WARNING: if nbTries is None, any infeasible condition shall cause an infinite loop
        '''
        n1 = 1 if n is None else n
        randomMCTuple = tuple(self.genRandomMC(n1,nbTries))
        if n is None:
            return randomMCTuple[0]
        return randomMCTuple
        
    def estimateMC(self,n,nbTries=None): 
        ''' returns an Alea instance, which is an estimation of the current distribution from a sample
            of n random values; this is a true Monte-Carlo algorithm, which does not precalculate the
            exact probability distribution (contrarily to 'random' method); 
            the method is suited for complex distributions, when calculation of exact probability
            distribution is intractable; the larger the value of n, the better the returned estimation;
            nbTries, if not None, defines the maximum number of trials in case a random value
            is incompatible with a condition; this happens only if the current Lea instance
            is (referring to) an Ilea or Blea instance, i.e. 'given' or 'buildCPT' methods;
            WARNING: if nbTries is None, any infeasible condition shall cause an infinite loop
        '''
        return Lea.fromSeq(self.randomMC(n,nbTries))
    
    def nbCases(self):
        ''' returns the number of atomic cases evaluated to build the exact probability distribution;
            this provides a measure of the complexity of the probability distribution 
        '''
        return sum(1 for vp in self._genVPs())
    
    def isTrue(self):
        ''' returns True iff the value True has probability 1;
                    False otherwise;
            raises exception if some value is not boolean
        '''
        (n,d) = self._p(True,checkValType=True) 
        return n == d

    def isFeasible(self):
        ''' returns True iff the value True has a non-null probability;
                    False otherwise;
            raises exception if some value is not boolean
        '''
        (n,d) = self._p(True,checkValType=True)
        return n > 0
        
    def asString(self,kind='/',nbDecimals=6,chartSize=100):
        ''' returns, after evaluation of the probability distribution self, a string
            representation of it;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value with its
            probability in a format depending of given kind, which is string among
            '/', '.', '%', '-', '/-', '.-', '%-'; 
            the probabilities are displayed as
            - if kind[0] is '/' : rational numbers "n/d" or "0" or "1"
            - if kind[0] is '.' : decimals with given nbDecimals digits
            - if kind[0] is '%' : percentage decimals with given nbDecimals digits
            - if kind[0] is '-' : histogram bar made up of repeated '-', such that
                                  a bar length of histoSize represents a probability 1 
            if kind[1] is '-', the histogram bars with '-' are appended after 
                               numerical representation of probabilities
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used
        '''        
        return self.getAlea().asString(kind,nbDecimals,chartSize)

    def __str__(self):
        ''' returns, after evaluation of the probability distribution self, a string
            representation of it;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value  with its
            probability expressed as a rational number "n/d" or "0" or "1";
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used;
            called on evalution of "str(self)" and "repr(self)"
        '''        
        return self.getAlea().__str__()

    __repr__ = __str__

    def asFloat(self,nbDecimals=6):
        ''' returns, after evaluation of the probability distribution self, a string
            representation of it;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value with its
            probability expressed as decimal with given nbDecimals digits;
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used;
        '''
        return self.getAlea().asFloat(nbDecimals)

    def asPct(self,nbDecimals=1):
        ''' returns, after evaluation of the probability distribution self, a string
            representation of it;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value with its
            probability expressed as percentage with given nbDecimals digits;
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used;
        '''
        return self.getAlea().asPct(nbDecimals)
    
    def histo(self,size=100):
        ''' returns, after evaluation of the probability distribution self, a string
            representation of it;
            it contains one line per distinct value, separated by a newline character;
            each line contains the string representation of a value with its
            probability expressed as a histogram bar made up of repeated '-',
            such that a bar length of given size represents a probability 1
            if an order relationship is defined on values, then the values are sorted by 
            increasing order; otherwise, an arbitrary order is used;
        '''
        return self.getAlea().histo(size)

    def getAlea(self):
        ''' returns an Alea instance representing the distribution after it has been evaluated;
            if self is an Alea instance, then it returns itself,
            otherwise the newly created Alea is cached : the evaluation occurs only for the first
            call; for successive calls, the cached Alea instance is returned, which is faster 
        '''
        if self._alea is None:
            self._alea = self.new()
        return self._alea

    def new(self):
        ''' returns a new Alea instance representing the distribution after it has been evaluated;
            if self is an Alea, then it returns a clone of itself (independent)
            note that the present method is overloaded in Alea class, to be more efficient
        '''
        return Alea.fromValFreqs(*self._genVPs())

    def cumul(self):
        ''' evaluates the distribution, then,
            returns a tuple with probability weights p that self <= value ;
            the sequence follows the order defined on values (if an order relationship is defined
            on values, then the tuples follows their increasing order; otherwise, an arbitrary
            order is used, fixed from call to call
            Note : the returned value is cached
        '''
        return self.getAlea().cumul()
        
    def invCumul(self):
        ''' evaluates the distribution, then,
            returns a tuple with the probability weights p that self >= value ;
            the sequence follows the order defined on values (if an order relationship is defined
            on values, then the tuples follows their increasing order; otherwise, an arbitrary
            order is used, fixed from call to call
            Note : the returned value is cached
        '''
        return self.getAlea().invCumul()
        
    def randomIter(self):
        ''' evaluates the distribution, then,
            generates an infinite sequence of random values among the values of self,
            according to their probabilities
        '''
        return self.getAlea()._randomIter
        
    def random(self,n=None):
        ''' evaluates the distribution, then, 
            if n is None, returns a random value with the probability given by the distribution
            otherwise, returns a tuple of n such random values
        '''
        if n is None:
            return self.getAlea().randomVal()
        return tuple(islice(self.randomIter(),n))

    def randomDraw(self,n=None,sorted=False):
        ''' evaluates the distribution, then,
            if n=None, then returns a tuple with all the values of the distribution, in a random order
                       respecting the probabilities (the higher count of a value, the most likely
                       the value will be in the beginning of the sequence)
            if n > 0,  then returns only n different drawn values
            if sorted is True, then the returned tuple is sorted
        '''
        return self.getAlea().randomDraw(n,sorted)

    def mutualInformation(self,other):
        ''' returns a float number representing the mutual information between self and other,
            expressed in bits
        '''
        other = Lea.coerce(other)
        return max(0.,self.entropy + other.entropy - Clea(self,other).entropy)

    def information(self):
        ''' returns a float number representing the information of self being true,
            expressed in bits (assuming that self is a boolean distribution)
            raises an exception if self is certainly false
        '''
        return self.informationOf(True)

    def informationOf(self,val):
        ''' returns a float number representing the information of given val,
            expressed in bits
            raises an exception if given val is impossible
        '''        
        (p,count) = self._p(val)
        if p == 0:
            raise Lea.Error("no information from impossible value")
        return log2(count/float(p))

    def lr(self,*hypLeas):
        ''' returns a float giving the likelihood ratio (LR) of an 'evidence' E,
            which is self, for a given 'hypothesis' H, which is the AND of given
            hypLeas arguments; it is calculated as 
                  P(E | H) / P(E | not H)
            both E and H must be boolean probability distributions, otherwise
            an exception is raised;
            an exception is raised also if H is certainly true or certainly false      
        '''
        return self.given(*hypLeas).lr()
    
from alea import Alea
from clea import Clea
from plea import Plea
from tlea import Tlea
from ilea import Ilea
from rlea import Rlea
from blea import Blea
from flea import Flea
from flea1 import Flea1
from flea2 import Flea2
from flea2a import Flea2a

# Lea constants representing certain values
Lea.true  = Lea.coerce(True)
Lea.false = Lea.coerce(False)
Lea.zero  = Lea.coerce(0)

# Lea convenience functions

X = Lea.cprod

def P(lea1):
    ''' returns a ProbFraction instance representing the probability for
        lea1 to be True, from 0/1 to 1/1;
        raises an exception if some value in the distribution is not boolean
        (this is NOT the case with lea1.p(True))
        this is a convenience function equivalent to lea1.P
    '''
    return lea1.P

def Pf(lea1):
    ''' returns the probability for lea1 to be True, as a floating point
        number, from 0.0 to 1.0;
        raises an exception if some value in the distribution is not boolean
        (this is NOT the case with lea1.pmf(True))
        this is a convenience function equivalent to lea1.Pf
    '''
    return lea1.Pf



'''
--------------------------------------------------------------------------------

    rlea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea
from alea import Alea
from toolbox import dict

class Rlea(Lea):
    
    '''
    Rlea is a Lea subclass, which instance has other Lea instances as values.
    '''
    
    __slots__ = ('_leaOfLeas','_factors')

    def __init__(self,leaOfLeas):
        Lea.__init__(self)
        self._leaOfLeas = leaOfLeas
        leaCounts = tuple((lea_,sum(p1 for (v,p1) in lea_._genVPs())) for (lea_,p2) in leaOfLeas._genVPs())
        pcount = 1
        for (lea1_,count) in leaCounts:
            pcount *= count
        self._factors = dict((lea_,pcount//count) for (lea_,count) in leaCounts)
         
    def _getLeaChildren(self):
        return (self._leaOfLeas,)

    def _clone(self,cloneTable):
        return Rlea(self._leaOfLeas.clone(cloneTable))    

    def _genVPs(self):
        for (lea1,p1) in self._leaOfLeas._genVPs():
            factor = self._factors[lea1]
            p1 *= factor
            for (v,p2) in lea1._genVPs():
                yield (v,p1*p2)

    def _genOneRandomMC(self):
        for leaArg in self._leaOfLeas._genOneRandomMC():
            for v in leaArg._genOneRandomMC():
                yield v



'''
--------------------------------------------------------------------------------

    tlea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea
from flea2 import Flea2

class Tlea(Lea):

    '''
    Tlea is a Lea subclass, which instance represents a probability distribution obtained by applying
    a given 2-ary function repeatedly on given Lea instances, a given number of times. It allows to
    avoid tedious typing or explcit loops; also, it makes the calculation faster by using a dichotomic
    algorithm.
    '''

    __slots__ = ('_op','_lea1','_nTimes')

    def __init__(self,op,lea1,nTimes=2):
        Lea.__init__(self)
        self._op = op
        self._lea1 = lea1
        self._nTimes = nTimes
        if nTimes <= 0:
            raise Lea.Error("times method requires a strictly positive integer")

    def _getLeaChildren(self):
        return (self._lea1,)

    def _clone(self,cloneTable):
        return Tlea(self._op,self._lea1.clone(cloneTable),self._nTimes)
    
    def _genVPs(self,nTimes=None):
        if nTimes is None:
            nTimes = self._nTimes
        lea1 = self._lea1.new()
        if nTimes == 1:
            return lea1._genVPs()
        # nTimes >= 2 : use dichotomic algorithm
        nTimes1 = nTimes // 2
        tlea = Tlea(self._op,lea1,nTimes1)
        alea = tlea.getAlea()
        # alea = tlea
        flea = Flea2(self._op,alea,alea.clone())
        if nTimes%2 == 1:
            # nTimes is odd : nTimes = 2*nTimes1 + 1
            # operate with one more lea1 on the current result 
            flea = Flea2(self._op,flea,self._lea1)
        return flea._genVPs()

    def _genOneRandomMC(self):
        for v in self._lea1._genOneRandomMC():
            pass
        for _ in range(self._nTimes-1):
            for v2 in self._lea1._genOneRandomMC():
                v = self._op(v,v2)
        yield v



'''
--------------------------------------------------------------------------------

    leapp_compiler.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

import license
from leapp_translator import LeappTranslator
from toolbox import input

PY_FILE_HEADER = '''
# ----------------------------------------------------------------------------- 
# %s
# ----------------------------------------------------------------------------- 
# file generated by leapp %s, from compilation of %s
# WARNING: any update will be overwritten at next generation!
# ----------------------------------------------------------------------------- 

'''[1:]

import os
try:
    # try to activate terminal history 
    import readline
except:
    # OS dependent failure -> no terminal history
    pass            

class LeappCompiler(object):
    
    ''' 
    LeappCompiler provides static methods to compile source leapp files into
    Python files (.py). Alternatively, the generated Python code can be directly
    executed, without being dumped in a .py file
    '''

    @staticmethod
    def compileLeaFile(leaFilename):
        f = open(leaFilename)
        leaMultilineStatement = f.read()
        f.close()
        pMultilineStatement = 'from lea import *\n\n'
        pMultilineStatement += LeappTranslator.getTarget00(leaMultilineStatement)
        return pMultilineStatement

    @staticmethod
    def execPythonStatement(pStatement,leaFilename):
        code = compile(pStatement+'\n',leaFilename,'exec')
        exec(code)
            
    @staticmethod
    def compileAndExecLeaFile(leaFilename):
        pMultilineStatement = LeappCompiler.compileLeaFile(leaFilename)
        LeappCompiler.execPythonStatement(pMultilineStatement,leaFilename)

    @staticmethod
    def compileAndWriteLeaFile(leaFilename,force=False):
        pMultilineStatement = LeappCompiler.compileLeaFile(leaFilename)
        pythonFilename = os.path.splitext(leaFilename)[0] + '.py'
        if force or not os.path.exists(pythonFilename):
            canWrite = True
        else:
            res = input("overwrite %s (y/n)? "%pythonFilename)
            canWrite = res.strip().lower() == 'y'
        if canWrite:         
            pMultilineStatement = PY_FILE_HEADER%(pythonFilename,license.VER,leaFilename) + pMultilineStatement
            f = open(pythonFilename,'w')
            f.write(pMultilineStatement)
            f.close()
            print("'%s' written"%pythonFilename)



'''
--------------------------------------------------------------------------------

    dlea.py

--------------------------------------------------------------------------------
Copyright 2013 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea

class Dlea(Lea):
    
    '''
    Dlea is a Lea subclass, which instance represents a probability distribution of the
    sequences of values obtained by a given number of draws without replacement from a
    given Lea instance.
    '''
    
    __slots__ = ('_lea1','_nbValues')

    def __init__(self,lea1,nbValues):
        if nbValues <= 0:
            raise Lea.Error("draw method requires a strictly positive integer")
        Lea.__init__(self)
        self._lea1 = lea1
        self._nbValues = nbValues

    def _reset(self):
        self._lea1.reset()

    def _clone(self,cloneTable):
        return Dlea(self._lea1.clone(cloneTable),self._nbValues)
    
    def _genVPs(self,nbValues=None):
        if nbValues is None:
            nbValues = self._nbValues
        if nbValues == 1:
            for (v,p) in self._lea1.genVPs():
                yield ((v,),p)
        else:     
            for (v,p) in self._lea1._genVPs():
                lea2 = self._lea1.clone()
                dlea = Dlea(lea2.given(lea2!=v).getAlea(),nbValues-1)
                for (vt,pt) in dlea._genVPs():
                    yield ((v,)+vt,p*pt)



'''
--------------------------------------------------------------------------------

    leapp_translator.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.Leapp

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

import string
from prob_fraction import ProbFraction
from toolbox import zip

# leapp internal configuration parameters, language-dependant
IDENTIFIER_CHARACTERS = string.ascii_letters + string.digits + '_'
LITERAL_CHARACTERS = IDENTIFIER_CHARACTERS + '#'
STRING_SEPARATORS = ('"',"'")
PRINT_SUFFIXES = '%/.-' + string.whitespace

CONTINUATION_LINE_CHARS2 = ','

class LeappTranslator(object):

    ''' 
    LeappTranslator provides static methods to translate a multiline 
    leapp statement into Python statement.
    The entry-point method is getTarget00.
    
    IMPORTANT NOTE:
    The present implementation of LeappTranslator does NOT use usual
    recommanded parsing techniques nor specialized modules.
    For this reason, it is admittedly VERY convoluted!
    One of the rationale is that Leapp does NOT intend to be a true progamming
    language; it merely adds some syntactic sugar on top of Python, to ease
    the usage of Lea. The Python fragments interleaved in Leapp syntax are
    not "understood" by Leapp translator, which simply copy them on output.
    This brutal approach is inadequate for the maintenance and is unsatisfying
    for the Leapp user who receives poor syntax error diagnosis.
    
    For these reasons, the present implementation is sentenced to a deep refactoring!

    '''
    
    class Error(Exception):
        pass

    @staticmethod
    def unindent(line):
        unindentedLine = line.strip()
        if len(unindentedLine) == 0:
            indentStr = ''
        else:   
            indentStr = line[:line.index(unindentedLine)]
        return (indentStr,unindentedLine)

    @staticmethod
    def getTarget00(rMultilineStatement):
        ''' 
        '''
        leaStatementLines = rMultilineStatement.split('\n')
        # merge lines that require it
        leaStatementLines1 = []
        leaStatementLineParts = []
        deindent = False
        for leaStatementLine in leaStatementLines:
            leaStatementLine = leaStatementLine.rstrip()
            if len(leaStatementLine) == 0:
                if len(leaStatementLineParts) > 0:
                    if len(leaStatementLineParts) > 0:
                        leaStatementLines1.append(''.join(leaStatementLineParts))
                        leaStatementLineParts = []
                    leaStatementLines1.append('')
            else:
                if deindent:
                    leaStatementLine = leaStatementLine.lstrip()           
                lastChar = leaStatementLine[-1]
                if lastChar == '\\':
                    leaStatementLine = leaStatementLine[:-1]
                leaStatementLineParts.append(leaStatementLine)
                if len(leaStatementLine) <= 1 or lastChar not in CONTINUATION_LINE_CHARS2:
                    leaStatementLines1.append(''.join(leaStatementLineParts))
                    leaStatementLineParts = []
                    deindent = False
                else:
                    deindent = True
        if len(leaStatementLineParts) > 0:
            leaStatementLines1.append(''.join(leaStatementLineParts))
        pStatementLines = []
        for leaStatementLine in leaStatementLines1:
            (identStr,unindentedLeaStatementLine) = LeappTranslator.unindent(leaStatementLine)
            pStatementLines.append(identStr+LeappTranslator.getTarget0(unindentedLeaStatementLine))
        return '\n'.join(pStatementLines)

    @staticmethod
    def getTarget0(sourceFragment):
        # assume sourceFragment is stripped both sides
        lenSourceFragment = len(sourceFragment)
        headTarget = ''
        tailTarget = ''
        if lenSourceFragment >= 1 and sourceFragment[0] == ':':
            # print statement
            if len(sourceFragment) <= 1:
                # empty
                return "print('')"
            prefixLength = 1
            printTypeCode = ' '
            secondChar = sourceFragment[1]
            if secondChar in PRINT_SUFFIXES:
                printTypeCode = secondChar
                prefixLength += 1
            if len(sourceFragment) >= 3:
                thirdChar = sourceFragment[2]
                if thirdChar == '-':
                    printTypeCode += '-'
                    prefixLength += 1
            sourceFragment = sourceFragment[prefixLength:]
            headTarget = 'print(('
            tailTarget = ''
            if printTypeCode == '.':
                tailTarget += ').asFloat())'
            elif printTypeCode == '%':
                tailTarget += ').asPct())'
            elif '-' in printTypeCode:
                tailTarget = '%s).asString("%s"))' % (tailTarget,printTypeCode)
            else:
                tailTarget += '))'
        return headTarget + LeappTranslator.getTarget0b(sourceFragment) + tailTarget  
        
    @staticmethod
    def getTarget0b(sourceFragment):
        sourceFragment = sourceFragment.strip()
        isBool = len(sourceFragment) >= 1 and (sourceFragment[0] == '@')
        if isBool:
            sourceFragment = sourceFragment[1:]
        target = LeappTranslator.getTarget1(sourceFragment)
        if isBool:
            target = '(' + target + ').p(True)'
        return target
        
    @staticmethod
    def getTarget1(sourceFragment):
        sourceFragments = []
        currentStringDelim = None
        chars = []
        for c in sourceFragment:
            if currentStringDelim is None:
                if c == '#':
                    break
                if c in STRING_SEPARATORS:
                    currentStringDelim = c
                    if len(chars) > 0:
                        sourceFragments.append(''.join(chars))
                    chars = []
                chars.append(c)
            elif c == currentStringDelim:
                currentStringDelim = None
                chars.append(c)
                sourceFragments.append(''.join(chars))
                chars = []
            else:
                chars.append(c)
        if len(chars) > 0:
            sourceFragments.append(''.join(chars))
        extractedStrings = ['']
        unstringedFragmentParts = []
        for sourceFragment in sourceFragments:
            if sourceFragment[0] in STRING_SEPARATORS:
                extractedStrings.append(sourceFragment)
                unstringedFragmentParts.append('#')
            else:
                unstringedFragmentParts.append(sourceFragment)
        unstringedFragment = ''.join(unstringedFragmentParts)
        unstringedTargetFragment = LeappTranslator.getTarget(unstringedFragment)
        fragmentParts = []
        targetFragment = ''.join(s1+s2 for (s1,s2) in zip(extractedStrings,unstringedTargetFragment.split('#')))
        return targetFragment
    
    @staticmethod
    def convertFraction(fractionExpression):
        fractionExpression = fractionExpression.strip()
        if '/' in fractionExpression or fractionExpression.endswith('%') \
         or fractionExpression.startswith('0') or fractionExpression.startswith('.'):
            probFraction = ProbFraction(fractionExpression)
            return "%s,%s" % (probFraction.numerator,probFraction.denominator)
        return fractionExpression

    @staticmethod
    def smartSplit(fragment,sep):
        levelParentheses = 0
        levelBraces = 0
        levelBrackets = 0
        parts = []
        chars = []
        chars2 = []
        lenSep = len(sep)
        nbMatchChars = 0
        sepC = sep[0]
        for c in fragment:
            if c == '(':
                levelParentheses += 1
            elif c == ')':
                levelParentheses -= 1
            elif c == '{':
                levelBraces += 1
            elif c == '}':
                levelBraces -= 1
            elif c == '[':
                levelBrackets += 1
            elif c == ']':
                levelBrackets -= 1
            if c == sepC and levelParentheses+levelBraces+levelBrackets == 0:
                chars2.append(c)
                nbMatchChars += 1
                if nbMatchChars == lenSep:
                    parts.append(''.join(chars))
                    del chars[:]
                    del chars2[:]
                    nbMatchChars = 0
                sepC = sep[nbMatchChars]
            else:
                if nbMatchChars > 0:
                    nbMatchChars = 0
                    chars += chars2
                    del chars2[:] 
                    sepC = sep[0]    
                chars.append(c)
            if levelParentheses<0 or levelBraces<0 or levelBrackets<0:
                raise LeappTranslator.Error('missing opening delimiter')
        if levelParentheses>0 or levelBraces>0 or levelBrackets>0:
            raise LeappTranslator.Error('missing closing pairing')
        if len(chars) > 0:
            parts.append(''.join(chars))
        return parts        

    @staticmethod
    def treatProbWeightExpression(dictExpression):        
        if ':' in dictExpression:
            # dictionary literal: enclose it in brackets
            distributionItems = tuple(LeappTranslator.smartSplit(item,':') for item in LeappTranslator.smartSplit(dictExpression,','))
            probFractions = tuple(ProbFraction(probWeightString) for (valStr,probWeightString) in distributionItems)
            probWeights = ProbFraction.getProbWeights(probFractions)
            newDictExpression = ','.join('(%s,%d)'%(LeappTranslator.getTarget(valStr),probWeight) \
                                 for ((valStr,probWeightString),probWeight) in zip(distributionItems,probWeights))
            return 'Lea.fromValFreqs(%s)' % newDictExpression
        # dictionary variable: expand items
        return 'Alea.fromValFreqsDictGen(%s)' % dictExpression
        
    @staticmethod
    def treatCPTExpression(cptExpression):
        if '->' in cptExpression:
            # CPT literal
            def f(condExpr):
                if condExpr.strip() == '_':
                    return None
                return condExpr    
            cptItems = tuple(LeappTranslator.smartSplit(item,'->') for item in LeappTranslator.smartSplit(cptExpression,','))
            newCPTExpression = ','.join('(%s,%s)'%(f(condExpr),distribExpr) for (condExpr,distribExpr) in cptItems)
            return newCPTExpression
        return cptExpression
        
    @staticmethod
    def getTarget(sourceFragment):
        targetFragment = sourceFragment
        targetFragment = LeappTranslator.parse(targetFragment,'?*','(',')','Lea.cprod(%s)')
        targetFragment = LeappTranslator.parse(targetFragment,'?!' ,'(',')','Lea.buildCPT(*(%s,))',LeappTranslator.treatCPTExpression)
        targetFragment = LeappTranslator.parse(targetFragment,'!!' ,'(',')','.revisedWithCPT(*(%s,))',LeappTranslator.treatCPTExpression)
        targetFragment = LeappTranslator.parse(targetFragment,'?' ,'(',')','Lea.fromVals(*(%s))')
        targetFragment = LeappTranslator.parse(targetFragment,'?:','(',')','Lea.boolProb(%s)',LeappTranslator.convertFraction)
        targetFragment = LeappTranslator.parse(targetFragment,'?' ,'{','}','%s',LeappTranslator.treatProbWeightExpression)
        targetFragment = LeappTranslator.parse(targetFragment,'!' ,'','','.given(%s)')
        targetFragment = LeappTranslator.parse(targetFragment,'$_' ,'(',')','.randomDraw(%s)')
        targetFragment = LeappTranslator.parse(targetFragment,'$' ,'(',')','.random(%s)')
        targetFragment = LeappTranslator.parseSingle(targetFragment,'$','.random()')
        targetFragment = LeappTranslator.parseSingle(targetFragment,'.@','.new()')
        targetFragment = LeappTranslator.parse(targetFragment,'@' ,'(',')','.p(%s)')
        targetFragment = LeappTranslator.parseIsolatedAt(targetFragment)
        targetFragment = LeappTranslator.parseSingle(targetFragment,'@','.p(True)')
        (head,timesArgs,identifier,tail) = LeappTranslator.parseIdentifier(targetFragment,'?')
        if head is not None:
            if identifier is None:
                raise LeappTranslator.Error('missing identifier')
            timesExpr = ''
            if timesArgs is not None:
                timesExpr = '.times(%s)'% timesArgs
            isFunc = tail.lstrip().startswith('(')
            if isFunc:
                targetFragment = head + LeappTranslator.parse(tail,'','(',')',('Flea.build(%s,(%%s,))'%identifier)+timesExpr)
            else:
                targetFragment = head + identifier
                if timesExpr:
                    targetFragment += timesExpr
                else:
                    targetFragment += '.clone()'
                targetFragment += LeappTranslator.getTarget(tail)
        return targetFragment

    @staticmethod
    def parseIsolatedAt(sourceFragment):
        splitResult = sourceFragment.split('@',1)
        if len(splitResult) == 1:
            return sourceFragment
        (head,tail) = splitResult
        if len(tail) == 0 or tail[0] not in LITERAL_CHARACTERS:
            return sourceFragment
        splitResult2 = tail.split(None,1)
        if len(splitResult2) == 2:
            (body,tail2) = splitResult2
        else:
            body = tail
            tail2 = ''
        return '%s.p(%s) %s' % (head,LeappTranslator.getTarget(body),LeappTranslator.getTarget(tail2))

    @staticmethod
    def parseSingle(sourceFragment,prefixToken,targetBody):
        splitResult = sourceFragment.split(prefixToken,1)
        if len(splitResult) == 1:
            return sourceFragment
        (head,tail) = splitResult
        return head + targetBody + LeappTranslator.getTarget(tail)

    @staticmethod
    def parse(sourceFragment,prefixToken,openToken,closeToken,targetBodyTemplate,treatFunc=None):
        special = openToken == ''
        if prefixToken == '':
            (head1,tail1) = ('',sourceFragment)
        else:
            splitResult1 = sourceFragment.split(prefixToken,1)
            if len(splitResult1) == 1:
                return sourceFragment
            (head1,tail1) = splitResult1
            #head1 = head1.rstrip()
        if special:
            if len(tail1) > 0 and tail1[0] == '=':
                # ignore Python inequality operator '!='
                return head1 + '!' + LeappTranslator.parse(tail1,prefixToken,openToken,closeToken,targetBodyTemplate)
            (head2,tail2) = ('',tail1)
            (openToken,closeToken) = ('(',')')
        else:
            splitResult2 = tail1.split(openToken,1)
            if len(splitResult2) == 1:
                return sourceFragment
            (head2,tail2) = splitResult2
        if len(head2.lstrip()) > 0:
            return sourceFragment
        bodyChars = []
        tailIter = iter(tail2)
        level = 1
        for c in tailIter:
            if c == openToken:
                level += 1
            elif c == closeToken:
                level -= 1
                if level == 0:
                    if special:
                        bodyChars.append(closeToken)
                    break
            bodyChars.append(c)
        if level > 0 and not special:
            raise LeappTranslator.Error("opening '%s' but missing '%s'"%(prefixToken+openToken,closeToken))
        body = ''.join(bodyChars)
        tail = ''.join(tailIter)
        if treatFunc is not None:
            body = treatFunc(body)
        targetBody = LeappTranslator.getTarget(body)
        return head1 + targetBodyTemplate % targetBody + LeappTranslator.getTarget(tail)
        
    @staticmethod
    def getIdentifier(sourceFragment):
        identifierChars = []
        tailIter = iter(sourceFragment)
        tail = ''
        for c in tailIter:
            if c not in IDENTIFIER_CHARACTERS:
                tail = c
                break
            identifierChars.append(c)
        identifier = ''.join(identifierChars)
        tail += ''.join(tailIter)
        return (identifier,tail)

    @staticmethod
    def parseIdentifier(sourceFragment,prefixToken):
        (head,timesArgs,identifier,tail) = (None,None,None,None)
        splitResult = sourceFragment.split(prefixToken,1)
        if len(splitResult) == 2:
            (head,tail1) = splitResult
            tail1 = tail1.lstrip()
            if len(tail1) > 0:
                firstChar = tail1[0]
                if firstChar == '[':
                    splitResult = tail1[1:].split(']',1)
                    if len(splitResult) == 2:
                        (timesArgs,tail1) = splitResult
                        tail1 = tail1.lstrip()
            if len(tail1) > 0:
                firstChar = tail1[0]
                if firstChar in string.ascii_letters:
                    (identifier,tail) = LeappTranslator.getIdentifier(tail1)
        return (head,timesArgs,identifier,tail)




'''
--------------------------------------------------------------------------------

    leapp.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

import sys
import license

''' 
Leapp main program
------------------
If no argument is given, then the Leapp console is open and the command loop is started.
Otherwise, the files refered by the given filename arguments are compiled into Python
statements. If '-c' is given, it writes the Python statements into .py files, otherwise,
it executes directly the Python statements.
'''

if __name__=='__main__':
    if len(sys.argv) == 1:
        # no argument : start leapp console
        from leapp_console import LeappConsole
        import platform
        leappConsole = LeappConsole()
        print ("[running on Python %s]"%platform.python_version())
        print (license.licenseText)
        print ('Welcome in Leapp console!')
        leappConsole.startCmdLoop()
        print ()
    else:
        # argument(s) present : compile files into Python statements then,
        # if '-c' given, write Python statements .py files
        # otherwise, execute Python statements 
        from leapp_compiler import LeappCompiler
        firstArg = sys.argv[1]
        if firstArg.startswith('-'):
            if firstArg not in ('-c','-cf'):
                print ("ERROR: unknown option '%s'"%firstArg)
                sys.exit(-1)                
        isCompile = firstArg.startswith('-c')
        force = firstArg == '-cf'
        if isCompile:
            leaFilenamesArgIdx = 2
        else:
            leaFilenamesArgIdx = 1
        leaFilenames = sys.argv[leaFilenamesArgIdx:]
        processLeaFile = LeappCompiler.compileAndWriteLeaFile if isCompile else LeappCompiler.compileAndExecLeaFile
        if isCompile:
            for leaFilename in leaFilenames:
                LeappCompiler.compileAndWriteLeaFile(leaFilename,force)
        else:
            for leaFilename in leaFilenames:
                LeappCompiler.compileAndExecLeaFile(leaFilename)



'''
--------------------------------------------------------------------------------

    blea.py

--------------------------------------------------------------------------------
Copyright 2013, 2014, 2015 Pierre Denis

This file is part of Lea.

Lea is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Lea is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with Lea.  If not, see <http://www.gnu.org/licenses/>.
--------------------------------------------------------------------------------
'''

from lea import Lea
from clea import Clea
from ilea import Ilea
from prob_fraction import ProbFraction
from toolbox import dict, genPairs
from operator import or_
from itertools import chain
            
class Blea(Lea):
    
    '''
    Blea is a Lea subclass, which instance represents a conditional probability
    table (CPT), as a set of Ilea instances. Each Ilea instance represents a
    given distibution <Vi,p(Vi|C)>, assuming a given condition C is verified,
    in the sense of a conditional probability.
    The set of conditions shall form a partition of the "certain true", i.e.
     ORing  all conditions shall give a "certain true" distribution
     ANDing all conditions pairwise shall give "certain false" distributions
    '''

    __slots__ = ('_ileas','_ctxClea','_condClea')
    
    def __init__(self,*ileas):
        Lea.__init__(self)
        self._ileas = tuple(ileas)
        # _condLea is used only by _genOneRandomMC method
        self._condClea = None
        # _ctxClea is a cartesian product of all Alea leaves present in CPT and having
        # multiple possible values; it is needed only if some clauses miss variables
        # present in other clauses (e.g. CPT with context-specific independence); a
        # rebalancing is needed if there are such missing variables and if these admit
        # multiple possible values (total probability weight > 1)
        aleaLeavesSet = frozenset(aleaLeaf for ilea in ileas                       \
                                           for aleaLeaf in ilea.getAleaLeavesSet() \
                                           if aleaLeaf._count > 1                  )
        self._ctxClea = Clea(*aleaLeavesSet)

    _argNamesOfBuildMeth = frozenset(('priorLea','autoElse','check'))

    @staticmethod
    def build(*clauses,**kwargs):
        argNames = frozenset(kwargs.keys())
        unknownArgNames = argNames - Blea._argNamesOfBuildMeth
        if len(unknownArgNames) > 0:
            raise Lea.Error("unknown argument keyword '%s'; shall be only among %s"%(next(iter(unknownArgNames)),tuple(Blea._argNamesOfBuildMeth)))
        priorLea = kwargs.get('priorLea',None)
        autoElse = kwargs.get('autoElse',False)
        check = kwargs.get('check',True)
        ## in PY3, could use: def build(*clauses,priorLea=None,autoElse=False,check=True):
        elseClauseResults = tuple(result for (cond,result) in clauses if cond is None)
        if len(elseClauseResults) > 1:
            raise Lea.Error("impossible to define more than one 'else' clause")
        if len(elseClauseResults) == 1:
            if priorLea is not None:
                raise Lea.Error("impossible to define together prior probabilities and 'else' clause")
            if autoElse:
                raise Lea.Error("impossible to have autoElse=True and 'else' clause")                
            elseClauseResult = elseClauseResults[0]
        elif autoElse:
            if priorLea is not None:
                raise Lea.Error("impossible to define together prior probabilities and autoElse=True")
            # take uniform distribution on all values found in clause's results (principle of indifference)
            elseClauseResult = Lea.fromVals(*frozenset(val for (cond,result) in clauses for val in Lea.coerce(result).vals()))
        else:
            elseClauseResult = None
        normClauseLeas = tuple((Lea.coerce(cond),Lea.coerce(result)) for (cond,result) in clauses if cond is not None)
        condLeas = tuple(condLea for (condLea,resultLea) in normClauseLeas)
        # check that conditions are disjoint
        if check:
            if any(v.count(True) > 1 for (v,_) in Clea(*condLeas)._genVPs()):
                raise Lea.Error("clause conditions are not disjoint")
        # build the OR of all given conditions
        orCondsLea = Lea.reduce(or_,condLeas,True)
        if priorLea is not None:
            # prior distribution: determine elseClauseResult
            if check and orCondsLea.isTrue():
                # TODO check priorLea equivalent to self
                raise Lea.Error("forbidden to define prior probabilities for complete clause set")
            (pTrue,count) = orCondsLea._p(True)
            pFalse = count - pTrue
            priorAleaDict = dict(priorLea.getAlea()._genVPs())
            priorAleaCount = sum(priorAleaDict.values())
            normAleaDict = dict(Lea.fromSeq(resultLea for (condLea,resultLea) in normClauseLeas).flat().getAlea()._genVPs())
            normAleaCount = sum(normAleaDict.values())
            valuesSet = frozenset(chain(priorAleaDict.keys(),normAleaDict.keys()))
            vps = []
            for value in valuesSet:
                 priorP = priorAleaDict.get(value,0)
                 condP = normAleaDict.get(value,0)
                 p = priorP*count*normAleaCount - condP*pTrue*priorAleaCount
                 if not(0 <= p <= pFalse*normAleaCount*priorAleaCount):
                     # Infeasible : probability represented by p goes outside range from 0 to 1
                     priorPFraction = ProbFraction(priorP,priorAleaCount)
                     lowerPFraction = ProbFraction(condP*pTrue,count*normAleaCount)
                     upperPFraction = ProbFraction(condP*pTrue+pFalse*normAleaCount,count*normAleaCount)
                     raise Lea.Error("prior probability of '%s' is %s, outside the range [ %s , %s ]"%(value,priorPFraction,lowerPFraction,upperPFraction))
                 vps.append((value,p))
            elseClauseResult = Lea.fromValFreqs(*vps)
        elif elseClauseResult is None:
            # no 'else' clause: check that clause set is complete
            if check and not orCondsLea.isTrue():
                raise Lea.Error("incomplete clause set requires 'else' clause or autoElse=True or priorLea=...")
        if elseClauseResult is not None:
            # add the else clause
            elseCondLea = ~orCondsLea
            ## other equivalent statement: elseCondLea = Lea.reduce(and_,(~condLea for condLea in condLeas))
            normClauseLeas += ((elseCondLea,Lea.coerce(elseClauseResult)),)
            # note that orCondsLea is NOT extended with orCondsLea |= elseCondLea
            # so, in case of else clause (and only in this case), orCondsLea is NOT certainly true
        return Blea(*(Ilea(resultLea,(condLea,)) for (condLea,resultLea) in normClauseLeas))    

    def _getLeaChildren(self):
        return self._ileas + (self._ctxClea,)
    
    def _clone(self,cloneTable):
        return Blea(*(iLea.clone(cloneTable) for iLea in self._ileas))

    def _genVPs(self):
        for iLea in self._ileas:
            for (v,p) in iLea._genVPs():
                for (_,p2) in self._ctxClea:
                    yield (v,p*p2)

    def _genOneRandomMC(self):
        if self._condClea is None:
            # _condAlea is a cartesian product of all Alea leaves present in CPT conditions;
            condAleaLeavesSet = frozenset(aleaLeaf for ilea in self._ileas                             \
                                                   for aleaLeaf in ilea._condLeas[0].getAleaLeavesSet())
            self._condClea = Clea(*condAleaLeavesSet)
        # the first for loop binds a random value on each Alea instances refered in CPT conditions
        for _ in self._condClea._genOneRandomMC():
            # here, there will be at most one ilea having condition that evaluates to True,
            # regarding the random binding that has been made 
            for iLea in self._ileas:
                for v in iLea._genOneRandomMCNoExc():
                    if v is not iLea:
                        # the current ilea is the one having the condition that evaluates to True
                        yield v
